from .environment import MultiTaskAmbient, MultiTaskEnvironment
from . import event
from . import agent
from . import extras

__all__ = (
    "MultiTaskEnvironment",
    "MultiTaskAmbient",
    "event",
    "agent",
    "extras",
)



"""Module that defines a parser for parsing event log files, see also `icua.utils.EventLogger`."""

from datetime import datetime
from pathlib import Path
import pkgutil
import importlib

from star_ray import Event


class EventLogParser:
    """Parser for event logs."""

    def __init__(self, event_classes: list[type] = None):
        """Constructor.

        Args:
            event_classes (list[type], optional): list of classes that may appear in the event log. Defaults to None.
        """
        super().__init__()
        self._event_cls_map = (
            {c.__name__: c for c in event_classes} if event_classes else {}
        )

    def discover_event_classes(self, module: str):
        """Automatically discover all classes that subclass `star_ray.Event`. This may be slow, so it may be preferable to provide these classes explicitly in the `EventLogParser` constructor (if they are known in advance).

        Note that if a class cannot be found during parsing an exception will be raised, if this happens, try calling this with the module that contains the missing event class.

        Args:
            module (str): module to import from (e.g. "icua")
        """
        for c in discover_subclasses(Event, module):
            self._event_cls_map[c.__name__] = c

    def parse(self, file_path: str | Path, relative_start: bool = True):
        """Parse an event log file. The expected format per line is: {TIMESTAMP} {EVENT}.

        Args:
            file_path (str | Path): path to the event log file
            relative_start (bool): where to normalise timestamps to be relative to the first log entry.

        Yields:
            tuple[float, star_ray.Event]: (timestamp, event)
        """
        with open(file_path) as file:
            file_iter = iter(file.readlines())
            (start_time, e) = self._parse_line(next(file_iter))
            yield (start_time, e)
            if not relative_start:
                start_time = 0
            for line in file_iter:
                (ts, e) = self._parse_line(line)
                yield (ts - start_time, e)

    def _parse_timestamp_to_ms(self, timestamp_str: str) -> float:  # noqa
        """Parse a str timestamp to get milliseconds."""
        timestamp = datetime.strptime(timestamp_str, "%Y-%m-%d-%H-%M-%S-%f")
        return timestamp.timestamp()

    def _parse_line(self, line: str):
        """Parse a line of the event log file, expected format: {TIMESTAMP} {EVENT}."""
        parts = line.split(" ", 2)
        if len(parts) != 3:
            raise ValueError(f"Malformed line: {line}")
        timestamp_str, class_name, data_str = parts
        timestamp = self._parse_timestamp_to_ms(timestamp_str)
        cls = self._event_cls_map.get(class_name, None)
        if cls is None:
            raise ValueError(f"Missing class: {class_name}")
        return timestamp, cls.model_validate_json(data_str)


def import_submodules(package_name: str):
    """Dynamically import all submodules of a package."""
    package = importlib.import_module(package_name)
    package_path = package.__path__

    for loader, module_name, is_pkg in pkgutil.walk_packages(
        package_path, package_name + "."
    ):
        importlib.import_module(module_name)


def find_all_subclasses(cls: type) -> list[type]:
    """Recursively find all subclasses of a given class."""
    subclasses = set(cls.__subclasses__())
    all_subclasses = set(subclasses)

    while subclasses:
        subclass = subclasses.pop()
        nested_subclasses = set(subclass.__subclasses__())
        subclasses.update(nested_subclasses)
        all_subclasses.update(nested_subclasses)

    return list(all_subclasses)


def discover_subclasses(base_class: type, package_name: str) -> list[type]:
    """Ensure all subclasses of a given base class are discovered within a package."""
    import_submodules(package_name)
    return list(sorted(find_all_subclasses(base_class), key=lambda c: c.__module__))



"""Module defining guidance related events."""

from typing import ClassVar
from pydantic import field_validator
import lxml.etree as etree
import math

from star_ray.event import Action
from star_ray_xml import (
    XMLState,
    XPathQuery,
    insert,
    update,
    select,
    XPathElementsNotFound,
)
from star_ray_pygame.cairosurface import parse_transform
from star_ray_pygame import SVGAmbient


__all__ = (
    "TaskAcceptable",
    "TaskUnacceptable",
    "ShowGuidance",
    "HideGuidance",
    "XPathAction",
    "DrawArrowAction",
    "DrawBoxAction",
    "DrawBoxOnElementAction",
    "ShowElementAction",
    "HideElementAction",
)


class TaskAcceptable(Action):
    """This action to be taken when a task goes from an `acceptable` state to an `unacceptable` state."""

    task: str

    def __execute__(self, state: XMLState):  # noqa
        pass


class TaskUnacceptable(Action):
    """This action to be taken when a task goes from an `acceptable` state to an `unacceptable` state."""

    task: str

    def __execute__(self, state: XMLState):  # noqa
        pass


class ShowGuidance(Action):
    """This action should be taken when guidance starts being shown by an agent."""

    task: str

    def __execute__(self, state: XMLState):  # noqa
        pass


class HideGuidance(Action):
    """This action should be taken when guidance stops being shown by an agent."""

    task: str

    def __execute__(self, state: XMLState):  # noqa
        pass


class XPathAction(XPathQuery):
    """Base class for guidance actions, validates the xpath attribute of XPathQuery."""

    # @validator("xpath", pre=True, always=True)
    @field_validator("xpath", mode="before")
    @classmethod
    def _validate_xpath(cls, xpath):
        if not isinstance(xpath, str):
            raise TypeError(f"Invalid xpath: '{xpath}', must be of type `str`")
        if xpath.endswith("/"):
            xpath = xpath[:-1]
        return xpath

    @property
    def is_read(self):  # noqa: D102
        return False

    @property
    def is_write(self):  # noqa: D102
        return True

    @property
    def is_write_tree(self):  # noqa: D102
        return True  # possibly

    @property
    def is_write_element(self):  # noqa: D102
        return True


class DrawElementAction(XPathAction):
    data: dict[str, str]

    REQUIRED_DATA: ClassVar[tuple[str]] = tuple()

    # @validator("data", pre=True, always=True)
    @field_validator("data", mode="before")
    @classmethod
    def _validate_data(cls, data):
        data = DrawElementAction._validate_none_data(data)
        return DrawElementAction._validate_required_data(data)

    @classmethod
    def _validate_none_data(cls, data):
        def _validate(data):
            for k, v in data.items():
                if v is None:
                    raise ValueError(
                        f"{DrawBoxAction.__name__} attribute: {k} cannot be `None`."
                    )
                yield str(k), str(v)

        return dict(_validate(data))

    @classmethod
    def _validate_required_data(cls, data):
        if not all(x in data for x in cls.REQUIRED_DATA):
            missing = [x for x in cls.REQUIRED_DATA if x not in data]
            raise ValueError(
                f"{DrawBoxAction.__name__} is missing attributes: {missing}"
            )
        return data

    @classmethod
    def _insert(cls, state: XMLState, data: dict[str, str], xpath: str):
        raise NotImplementedError()

    @classmethod
    def _update(cls, state: XMLState, data: dict[str, str], xpath: str):
        raise NotImplementedError()

    @classmethod
    def _draw(cls, state: XMLState, data: dict[str, str], xpath: str):
        raise NotImplementedError()

    def __execute__(self, state: XMLState):
        return type(self)._draw(state, self.data, self.xpath)


class DrawArrowAction(DrawElementAction):
    """Event that will draw an arrow in the UI.

    Optional values: scale, id, x, y, opacity, rotation, fill, stroke, stroke_width.
    """

    ARROW_SVG: ClassVar[str] = (
        """<svg:svg xmlns:svg="http://www.w3.org/2000/svg" transform="scale({scale})" id="{id}" x="{x}" y="{y}" opacity="{opacity}" viewBox="-28 -28 178 178" width="200" height="200"><svg:polygon transform="rotate({rotation}, 61.44, 61.44)" fill="{fill}" stroke="{stroke}" stroke-width="{stroke_width}" points="148.33, 61.44 32.07, 118.96 62.71, 61.44 32.07, 3.92 148.33, 61.44" /> </svg:svg>"""
    )
    REQUIRED_DATA: ClassVar[tuple[str]] = ("id",)

    @classmethod
    def new_rotation(cls, rotation: float) -> str:
        """TODO this should be a utility method."""
        return f"rotate({rotation}, 61.44, 61.44)"

    @classmethod
    def new_scale(cls, scale: float) -> str:
        """TODO this should be a utility method."""
        return f"scale({scale})"

    @classmethod
    def _insert(cls, state: XMLState, data: dict[str, str], xpath: str):
        data.setdefault("fill", "none")
        data.setdefault("opacity", "1")
        data.setdefault("stroke", "#ff0000")
        data.setdefault("scale", "1")
        data.setdefault("rotation", "0")
        data.setdefault("stroke_width", "2.0")
        data.setdefault("x", "0")
        data.setdefault("y", "0")
        element = DrawArrowAction.ARROW_SVG.format(**data)
        return state.insert(insert(xpath, element, index=10))

    @classmethod
    def _set_if(cls, attr: str, element: etree.ElementBase, data: dict[str, str]):
        if attr in data:
            element.set(attr, data[attr])

    @classmethod
    def _update(cls, state: XMLState, data: dict[str, str], xpath: str):
        pol_data = {
            k.replace("_", "-"): data[k]
            for k in ("fill", "stroke", "stroke_width")
            if k in data
        }
        if "scale" in data:
            transform = DrawArrowAction.new_scale(data["scale"])
            state.update(update(xpath=xpath, attrs={"transform": "transform"}))

        # update x and y, using x and y as the center coordinates
        if "x" in data or "y" in data:
            x, y = data.get("x", None), data.get("y", None)
            size, _ = DrawArrowAction.get_size_scale(state, xpath)
            attrs = {k: data[k] for k in ("opacity",) if k in data}
            if x is not None:
                attrs["x"] = str(float(x) - size[0] / 2)
            if y is not None:
                attrs["y"] = str(float(y) - size[1] / 2)
            state.update(update(xpath=xpath, attrs=attrs))

        if "rotation" in data:
            pol_data["transform"] = DrawArrowAction.new_rotation(data["rotation"])
        pol_xpath = xpath + "/svg:polygon"
        state.update(update(xpath=pol_xpath, attrs=pol_data))

        # check if we need to update the arrow to point to a target
        point_to = data.get("point_to", None)
        if point_to:
            rotation = DrawArrowAction.rotation_from_point_to(state, point_to, xpath)
            transform = DrawArrowAction.new_rotation(rotation)
            state.update(update(xpath=pol_xpath, attrs={"transform": transform}))

    @classmethod
    def _draw(cls, state: XMLState, data: dict[str, str], xpath: str):
        _id = data.get("id", None)
        assert _id is not None  # must have an 'id'
        # does the element already exist?
        uxpath = xpath + f"/svg:svg[@id='{data['id']}']"
        try:
            state.select(select(xpath=uxpath, attrs=["id"]))
        except XPathElementsNotFound:
            # it doesnt exist, create it
            return cls._insert(state, data, xpath)
        # it already exists, update it
        cls._update(state, data, uxpath)

    @staticmethod
    def get_size_scale(state: XMLState, xpath: str):
        """TODO this should be a utility method."""
        result = state.select(
            select(xpath, attrs=["x", "y", "width", "height", "transform"])
        )
        if not result:
            raise ValueError(f"Element at xpath: {xpath} doesn't exist")
        values = result[0]
        scale, _, _ = parse_transform(values["transform"])
        return (values["width"] * scale[0], values["height"] * scale[1]), scale

    @staticmethod
    def get_element_center(state: XMLState, xpath: str):
        """TODO this should be a utility method."""
        result = state.select(
            select(xpath, attrs=["x", "y", "width", "height", "transform"])
        )
        if not result:
            raise ValueError(f"Element at xpath: {xpath} doesn't exist")
        values = result[0]
        x, y = (values["x"], values["y"])
        scale, _, _ = parse_transform(values["transform"])
        width, height = (values["width"] * scale[0], values["height"] * scale[1])
        return x + width / 2, y + height / 2

    @staticmethod
    def rotation_from_point_to(state: XMLState, element_id: str, xpath: str) -> float:
        """Computes a rotation that points from the center of the arrow to the center of the given element.

        Args:
            state (XMLState): state of the environment
            element_id (str): element to point to
            xpath (str): xpath of the arrow element.

        Returns:
            float: rotation
        """
        # TODO this should be a utility method somewhere that will compute the rotation between two elements given their xpaths!
        # get the center of the arrow
        (x1, y1) = DrawArrowAction.get_element_center(state, xpath)
        # get the center of the element with id `element_id`
        (x2, y2) = DrawArrowAction.get_element_center(
            state, f".//*[@id='{element_id}']"
        )
        # compute angle between points
        return math.degrees(math.atan2(y2 - y1, x2 - x1))


class ShowElementAction(XPathAction):
    """Action to show an element by setting its opacity to 1."""

    def __execute__(self, state: XMLState):  # noqa
        return state.update(update(self.xpath, attrs={"opacity": 1}))


class HideElementAction(XPathAction):
    """Action to hide an element by setting its opacity to 0."""

    def __execute__(self, state: XMLState):  # noqa
        return state.update(update(self.xpath, attrs={"opacity": 0}))


class DrawBoxAction(XPathAction):
    """TODO."""

    box_data: dict[str, str]

    # @validator("box_data", pre=True, always=True)
    @field_validator("box_data", mode="before")
    @classmethod
    def _validate_box_data(cls, data):
        data = DrawBoxAction._validate_none_box_data(data)
        return DrawBoxAction._validate_required_box_data(data)

    @classmethod
    def _validate_none_box_data(cls, data):
        def _validate(data):
            for k, v in data.items():
                if v is None:
                    raise ValueError(
                        f"{DrawBoxAction.__name__} attribute: {k} cannot be None"
                    )
                yield str(k), str(v)

        return dict(_validate(data))

    @classmethod
    def _validate_required_box_data(cls, data: dict[str, str]):
        if not all(x in data for x in ("x", "y", "width", "height")):
            missing = [x for x in ("x", "y", "width", "height") if x not in data]
            raise ValueError(
                f"{DrawBoxAction.__name__} is missing attributes: {missing}"
            )
        return data

    @staticmethod
    def _new_box(state: XMLState, box_data: dict[str, str], xpath: str):
        box_data.setdefault("stroke-width", "2")
        box_data.setdefault("stroke", "#ff0000")
        box_data.setdefault("fill", "none")
        box = etree.Element(
            f"{{{SVGAmbient.DEFAULT_SVG_NAMESPACES['svg']}}}rect",
            attrib=box_data,
            nsmap=SVGAmbient.DEFAULT_SVG_NAMESPACES,
        )
        box = etree.tostring(box)
        return state.insert(insert(xpath=xpath, element=box, index=0))

    @staticmethod
    def _draw_box(state: XMLState, box_data: dict[str, str], xpath: str):
        box_id = box_data.get("id", None)
        if box_id is None:
            raise ValueError("Attempted to draw a box without an `id` attribute.")
        # check if the box already exists
        uxpath = xpath + f"/svg:rect[@id='{box_id}']"
        try:
            state.select(select(xpath=uxpath, attrs=["id"]))
        except XPathElementsNotFound:
            # create a new box (it doesnt exist yet)
            return DrawBoxAction._new_box(state, box_data, xpath)
        return state.update(update(xpath=uxpath, attrs=box_data))

    def __execute__(self, state: XMLState):  # noqa
        return DrawBoxAction._draw_box(state, self.box_data, self.xpath)


class DrawBoxOnElementAction(DrawBoxAction):
    """TODO."""

    # @validator("box_data", pre=True, always=True)
    @field_validator("box_data", mode="before")
    @classmethod
    def _validate_box_data(cls, data):
        return DrawBoxAction._validate_none_box_data(data)

    def __execute__(self, state: XMLState):  # noqa
        required_attrs = ("x", "y", "width", "height")
        result = state.select(select(xpath=self.xpath, attrs=required_attrs))
        if not result:
            raise ValueError(
                f"Failed to find element at xpath: {self.xpath} for box draw."
            )
        if len(result) > 1:
            raise ValueError(
                f"Found multiple elements at xpath: {self.xpath} for box draw."
            )
        try:
            result = DrawBoxAction._validate_none_box_data(result[0])
            box_data = {**self.box_data}
            box_data.update(result)
            box_data = DrawBoxAction._validate_required_box_data(box_data)
        except ValueError as e:
            raise ValueError(
                f"Failed to validate action {DrawBoxOnElementAction.__name__} for element at xpath: {self.xpath}"
            ) from e
        DrawBoxAction._draw_box(state, box_data, self.xpath + "/parent::*")



import logging
import pathlib
from datetime import datetime
import time
from pydantic import BaseModel
from star_ray.pubsub import Subscriber


from loguru import logger as _logger
from pathlib import Path
import os
import sys

_CWD_DIRECTORY = Path(os.getcwd())
# TODO remove this...
_PROJECT_ROOT_DIRECTORY = "C:/Users/brjw/Documents/repos/dicelab"


def format_record(record):
    file_path = Path(record["file"].path)
    try:
        file_path = file_path.relative_to(_CWD_DIRECTORY)
    except ValueError:
        pass

    # Format the log message
    return "{level} | {file}:{line} | {message}\n".format(
        # time=record["time"].strftime("%H:%M:%S"),
        level=record["level"].name,
        file=str(file_path),
        line=record["line"],
        message=record["message"],
    )


# Configure the logger
LOGGER = _logger.bind(package="icua")


def set_level(level: str):
    LOGGER.remove()
    LOGGER.add(
        sink=sys.stdout,
        format=format_record,
        level=level,
    )


LOGGER.set_level = set_level

# default level
LOGGER.set_level("DEBUG")

__all__ = ("LOGGER",)


class DateTimeFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        if not datefmt:
            return super().formatTime(record)
        return datetime.fromtimestamp(record.created).strftime(datefmt)


class EventLogger(Subscriber):
    def __init__(self, path: str = None):
        path = path if path is not None else EventLogger.default_log_path()
        if isinstance(path, str):
            path = pathlib.Path(path).expanduser().resolve()
        pathlib.Path(path).parent.mkdir(exist_ok=True)
        self.path = str(path)
        self.logger = logging.getLogger(f"{path.name.split('.')[0]}_event_logger")
        self.logger.setLevel(logging.INFO)
        file_handler = logging.FileHandler(self.path)
        # TODO REMOVE DEPRECATED - time stamp is now set in log to be more in line with event execution times
        # formatter = DateTimeFormatter(
        #     fmt="%(asctime)s %(message)s", datefmt="%Y-%m-%d-%H-%M-%S-%f"
        # )
        # file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)

    def log(self, event: BaseModel):
        t = datetime.fromtimestamp(time.time()).strftime("%Y-%m-%d-%H-%M-%S-%f")
        self.logger.info("%s %s %s", t, type(event).__name__, event.model_dump_json())

    def __notify__(self, event: BaseModel):
        self.log(event)

    @staticmethod
    def default_log_path(path=None, name=None):
        if path is None:
            path = "./logs/"
        current_time = datetime.now()
        timestamp = current_time.strftime("%Y-%m-%d-%H-%M-%S")
        name = name if name is not None else f"event_log_{timestamp}.log"
        if "{datetime}" in name:
            name = name.format(datetime=timestamp)
        default_filename = pathlib.Path(path, name).expanduser().resolve()
        return default_filename



"""Module defines some common eyetracking filters that can be used with an `Eyetracker`.

Filter classes:
- `NWMAFilter` (Non-weighted Moving Average Filter)
- `IVTFilter` (Velocity-Threshold Identification Filter)
- `WindowSpaceFilter`

See class documentation for details.
"""
# TODO test the IVTFilter using mouse? It is very important that this works correctly! we can also find a good value for it by introducing some noise into the mouse position

from typing import Any
from collections import deque
import math
from ...utils import LOGGER

__all__ = ("NWMAFilter", "WindowSpaceFilter", "IVTFilter")


class NWMAFilter:  # Non-weighted moving average
    """Filter that computes a moving average over eye gaze positions for some window size."""

    def __init__(self, n: int):
        """Constructor.

        Args:
            n (int): window size of the moving average.
        """
        super().__init__()
        self.data_x = deque(maxlen=n)
        self.data_y = deque(maxlen=n)

    def __call__(self, data: dict[str, Any]) -> dict[str, Any]:
        """Compute the moving average. Expects a `position` attribute in `data` and will modify this attribute in-place.

        Args:
            data (dict[str, Any]): eyetracking data.

        Returns:
            dict[str, Any]: filtered eyetracking data.
        """
        x, y = data["position"]
        # print("nwma:", x, y)
        if math.isnan(x) or math.isnan(y):
            return data  # ignore this sample
        self.data_x.append(x)
        self.data_y.append(y)
        x = sum(self.data_x) / len(self.data_x)
        y = sum(self.data_y) / len(self.data_y)
        data["position"] = (x, y)
        return data

    def __str__(self) -> str:  # noqa
        return f"{NWMAFilter}(N={len(self.data_x)})"

    def __repr__(self) -> str:  # noqa
        return str(self)


class WindowSpaceFilter:
    """Filter that will transform eye gaze positions from screen space to window coordinate space."""

    def __init__(
        self,
        screen_size: tuple[float, float],
        window_size: tuple[float, float],
        window_position: tuple[float, float],
        keep_screen_data: bool = True,
    ):
        """Constructor.

        Args:
            screen_size (tuple[float, float]): size of the screen/computer monitor.
            window_size (tuple[float, float]): size of the UI window.
            window_position (tuple[float, float]): position of the UI window on the screen.
            keep_screen_data (bool): whether to preserve the screen data (position/velocity). If True then the fields `position_screen` and `velocity_screen` will be added, otherwise this information will be lost as `position` and `velocity` are modified in place.
        """
        super().__init__()
        self.screen_size = screen_size
        self.window_size = window_size
        self.window_position = window_position
        self._keep_screen_data = keep_screen_data

    def __call__(self, data: dict[str, Any]) -> dict[str, Any]:
        """Compute the screen->window space transformation. Expects a `position` attribute in `data` and will modify this attribute in-place. It will also transform the `velocity` attribute if it is found.

        This filter expects the screen coordinates to be in the range (0,1). It also assumes that the screen and the window share the same origin (0,0) being the top-left. The screen->window transformation is computed as:
        ```
        x *= screen_width
        y *= screen_height
        x -= window_x
        y -= window_y
        ```

        which is the default for many eyetrackers.
        If `keep_screen_data` is True then new attributes: `position_screen` and `position_velocity` will be added to preserve the screen position/velocity.

        Args:
            data (dict[str, Any]): eyetracking data.

        Returns:
            dict[str, Any]: filtered eyetracking data.
        """
        x, y = data["position"]
        # print("window:", x, y)
        if math.isnan(x) or math.isnan(y):
            data["in_window"] = False
            if self._keep_screen_data:
                data["position_screen"] = data["position"]
                if "velocity" in data:
                    data["velocity_screen"] = data["velocity"]
            return data

        position_screen = (x, y)
        if not (0 <= x <= 1 and 0 <= y <= 1.0):
            LOGGER.warning(
                f"Expected eyetracking coordinates in the range [0-1], received: {position_screen}."
            )
        x *= self.screen_size[0]
        y *= self.screen_size[1]
        x -= self.window_position[0]
        y -= self.window_position[1]
        in_window = (
            x > 0 and x < self.window_size[0] and y > 0 and y < self.window_size[1]
        )
        # print((x, y), self.screen_size, self.window_size)
        data["in_window"] = in_window
        data["position"] = (x, y)
        if self._keep_screen_data:
            data["position_screen"] = position_screen
        if "velocity" in data:
            vx, vy = data["velocity"]
            data["velocity"] = (vx * self.screen_size[0], vy * self.screen_size[1])
            if self._keep_screen_data:
                data["velocity_screen"] = (vx, vy)
        return data


class IVTFilter:
    """Velocity-Threshold Identification Filter (loosely based on http://www.vinis.co.kr/ivt_filter.pdf).

    This will filter will attempt to determine whether the eyes are fixated or saccading based on their velocity. The implementation computes positional velocity rather than anglular velocity (which is a departure from the original algorithm). The reason is that gaze angle may not be avaliable in the eyetracking event data.

    The `fixated` and `velocity` attributes will be added to eyetracking event data. `fixated` will be True if the velocity is below the threshold and False otherwise (indicating a saccade).

    IMPORTANT: The `velocity_threshold` is there for sensitive to the coordinate system that is in use, so be aware when using this along side coordinate space transformation filters.
    """

    def __init__(self, velocity_threshold: float):
        """Constructor.

        Args:
            velocity_threshold (float): velocity threshold used to determine whether the eye is fixated or saccading.
        """
        self.data_x = deque(maxlen=2)
        self.data_y = deque(maxlen=2)
        self.data_t = deque(maxlen=2)
        self.velocity_threshold = velocity_threshold

    def __call__(self, data: dict[str, Any]) -> dict[str, Any]:
        """Computes fixation/saccade based on velocity threshold. The velocity is estimated based on the two most recent positions and their associated timestamps. The first event will always be treated as a fixation. The filter assumes that the `position` and `timestamp` attribute are present in `data`.

        The `fixated` and `velocity` attributes will be added to eyetracking event data. `fixated` will be True if the velocity is below the threshold and False otherwise (indicating a saccade).

        Args:
            data (dict[str, Any]): eyetracking data.

        Returns:
            dict[str, Any]: filtered eyetracking data.
        """
        x, y = data["position"]
        t = data["timestamp"]
        # print("ivt:", x, y)

        if math.isnan(x) or math.isnan(y):
            data["fixated"] = False
            data["velocity"] = (float("nan"), float("nan"))
            return data  # ignore the data point

        # estimate velocity using the two most recent samples...
        self.data_x.append(x)
        self.data_y.append(y)
        self.data_t.append(t)
        dt = self.data_t[-1] - self.data_t[0]
        if dt > 0:
            dx = self.data_x[-1] - self.data_x[0]
            dy = self.data_y[-1] - self.data_y[0]
            speed = math.sqrt(dx**2 + dy**2) / dt
            data["velocity"] = (dx / dt, dy / dt)
        elif dt == 0:
            speed = 0
            data["velocity"] = (0, 0)
        else:
            raise ValueError(f"Negative dt: {dt} was found in {IVTFilter.__name__}.")
        data["fixated"] = speed <= self.velocity_threshold
        return data



"""Package containing extra functionality that may be useful in icua experiments.

Extras:
    - eyetracking: eyetracking support
    - analysis: tools for post-experiment analysis
"""

from . import analysis
from . import eyetracking

__all__ = ("eyetracking", "analysis")



from typing import Any
from collections.abc import Callable
import importlib.util
import inspect
import sys
import random
from lxml.etree import canonicalize
from jinja2 import Template
from functools import wraps
from pathlib import Path

from star_ray.utils import ValidatedEnvironment, TemplateLoader
from star_ray.agent import Agent, Actuator
from ._logging import LOGGER
from ._error import TaskConfigurationError
from ._schedule import ScheduledAgentFactory

__all__ = ("avatar_actuator", "agent_actuator", "TaskLoader")

EXT_SCHEMA = ".schema.json"
EXT_CONTEXT = ".json"
EXT_SCHEDULE = ".sch"
EXT_SVG = ".svg"
EXT_SVG_TEMPLATE = ".svg.jinja"
EXT_PY = ".py"

CLSATTR_IS_AVATAR_ACTUATOR = "__is_avatar_actuator__"
CLSATTR_IS_AGENT_ACTUATOR = "__is_agent_actuator__"


def avatar_actuator(cls: type[Actuator]):
    if not issubclass(cls, Actuator):
        raise TaskConfigurationError(
            f"Invalid use of @avatar, {cls} must derive `{Actuator.__name__}`"
        )
    setattr(cls, CLSATTR_IS_AVATAR_ACTUATOR, True)
    return cls


def agent_actuator(cls: type[Actuator]):
    if not issubclass(cls, Actuator):
        raise TaskConfigurationError(
            f"Invalid use of @agent, {cls} must derive `{Actuator.__name__}`"
        )
    setattr(cls, CLSATTR_IS_AGENT_ACTUATOR, True)
    return cls


class AvatarFactory:
    def __init__(self, actuators: list[type[Actuator] | Callable[[], Actuator]]):
        self._actuators = actuators

    def __call__(self, avatar: Agent):
        actuator_types = set(type(a) for a in avatar.get_actuators())
        for actuator in self._actuators:
            if actuator in actuator_types:
                raise ValueError(f"Avatar already has actuator of type: {actuator}.")
            avatar.add_component(actuator())
        return avatar


class Task:
    def __init__(
        self,
        task_name: str,
        task_template: Template,
        avatar_factory: Callable[[Agent], Agent],
        agent_factory: Callable[[], Agent],
    ):
        super().__init__()
        self._task_name = task_name
        self._task_template = task_template
        self._avatar_factory = avatar_factory
        self._agent_factory = agent_factory

    @property
    def task_name(self):
        return self._task_name

    def get_xml(self, context: dict[str, Any] | None = None) -> str:
        if context is None:
            context = dict()
        source = self._task_template.render(context)
        return canonicalize(source)

    def get_avatar(self, avatar: Agent) -> Agent:
        return self._avatar_factory(avatar)

    def get_agent(self) -> Agent:
        return self._agent_factory()


class _DefaultFuncs:
    @staticmethod
    def min(*args):
        return min(args)

    @staticmethod
    def max(*args):
        return max(args)

    @staticmethod
    def uniform(a: int | float, b: int | float):
        if isinstance(a, int) and isinstance(b, int):
            return random.randint(a, b)
        return random.uniform(a, b)


class TaskLoader:
    def __init__(self):
        super().__init__()
        self._jinja_env = ValidatedEnvironment(
            loader=TemplateLoader(),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def get_schedule_functions(self):
        return [
            _DefaultFuncs.min,
            _DefaultFuncs.max,
            _DefaultFuncs.uniform,
        ]

    def register_task(self, name: str, path: str | list[str]) -> None:
        if isinstance(path, str | Path):
            path = [TaskLoader._path_normalise(path)]
        elif isinstance(path, list | tuple):
            path = [TaskLoader._path_normalise(p) for p in path]
        LOGGER.debug(f"Registering task: `{name}` at path(s): `{[p for p in path]}`")
        self._template_loader.add_namespace(name, path)

    def load(
        self,
        task_name: str,
        avatar_actuators: list[Actuator],
        agent_actuators: list[Actuator],
    ) -> Task:
        task_template = self.get_task_template(task_name=task_name)
        agent_factory = self.get_schedule(task_name, agent_actuators)
        avatar_factory = AvatarFactory(avatar_actuators)
        return Task(
            task_name=task_name,
            task_template=task_template,
            avatar_factory=avatar_factory,
            agent_factory=agent_factory,
        )

    def get_task_template(self, task_name: str) -> Template:
        files = self._get_task_files(task_name)
        state_path, _, _ = self._validate_state_files(files, task_name=task_name)
        return self._jinja_env.get_template(state_path.as_posix())

    def get_task_source(
        self, task_name: str, context: dict[str, Any] | None = None
    ) -> str:
        if context is None:
            context = dict()  # no additional context
        template: Template = self.get_task_template(task_name)
        return template.render(context)

    def get_schedule(
        self, task_name: str, actuators: list[type[Actuator]]
    ) -> Callable[[], Agent]:
        actuators = set(actuators)
        has_actuators = len(actuators) > 0
        files = self._get_task_files(task_name)
        schedule_path = files.get(EXT_SCHEDULE, None)
        if schedule_path:
            # TODO it would be nice if this could give the full path, its may be useful to see exactly which file is being loaded
            LOGGER.debug(f"loading schedule: {schedule_path.name}")
            # TODO perhaps we shouldnt load as a template... just load as text?
            # the templating syntax might interfere with things?
            source = self._jinja_env.get_template(schedule_path.as_posix()).render()
            agent_factory = ScheduledAgentFactory(
                source,
                actuators,
                self.get_schedule_functions(),
            )
            return agent_factory
        elif has_actuators:
            raise TaskConfigurationError(
                f"Configuration file: `{task_name}{EXT_SCHEDULE}` is missing for task: `{task_name}` but actuators: {actuators} were specified."
            )

    def _validate_state_files(
        self, files: dict[str, str], task_name: str
    ) -> tuple[Path, Path, Path]:
        state_path, context_path, schema_path = None, None, None
        if EXT_SVG in files:
            state_path = files[EXT_SVG]
            LOGGER.debug(f"loading state: `{state_path.name}`")
        elif EXT_SVG_TEMPLATE in files:
            state_path = files[EXT_SVG_TEMPLATE]
            LOGGER.debug(f"loading state: `{ state_path.name}` ")
            has_schema = EXT_SCHEMA in files
            has_context = EXT_CONTEXT in files
            if not has_context and not has_schema:
                raise TaskConfigurationError(
                    f"Configuration file: `{task_name}{EXT_CONTEXT}` is missing from task template: `{state_path.name}`."
                )
            elif not has_schema:
                LOGGER.warning(
                    f"validation schema: `{task_name}{EXT_SCHEMA}` is missing from task template: `{state_path.name}`",
                )

            if has_schema:
                schema_path = files[EXT_SCHEMA]
                LOGGER.debug(
                    f"with validator schema: `{schema_path.name}`",
                )
            if has_context:
                context_path = files[EXT_CONTEXT]
                LOGGER.debug(
                    f"with context: `{context_path.name}`",
                )
        else:
            raise TaskConfigurationError(
                f"State file: `{task_name}{EXT_SVG}(.jinja)` is missing for task: `{task_name}`."
            )
        return state_path, context_path, schema_path

    def _get_task_files(self, task_name: str):
        """Generator that gets all file paths associated with the given task."""

        def _gen():
            for file in self._template_loader.list_templates_in_namespace(task_name):
                file = task_name / Path(file)
                if file.stem.startswith(task_name):
                    yield "".join(file.suffixes), file

        return dict(_gen())

    @staticmethod
    def _path_normalise(path: str | Path) -> Path:
        return Path(path).expanduser().resolve().absolute().as_posix()

    @property
    def _template_loader(self) -> TemplateLoader:
        return self._jinja_env.loader

    # @LOGGER.indent
    # def load_actuators(
    #     self,
    #     task_name: str,
    #     paths: List[str],
    #     agent_actuators: List[Type[Actuator]] | None,
    #     avatar_actuators: List[Type[Actuator]] | None,
    #     enable_dynamic_loading: bool = False,
    #     suppress_warnings: bool = False,
    # ):
    #     LOGGER.debug("loading actuators:")
    #     if agent_actuators is None:
    #         agent_actuators = []
    #     if avatar_actuators is None:
    #         avatar_actuators = []
    #     if enable_dynamic_loading:
    #         for path in paths:
    #             actuator_classes, _ = load_task_package_from_path(
    #                 path, task_name, suppress_warnings=suppress_warnings
    #             )
    #             if not actuator_classes["agent"] and not actuator_classes["avatar"]:
    #                 raise TaskConfigurationError(
    #                     f"No actuator classes were found in task plugin, did you forget to tag with @{agent_actuator.__name__} or @{avatar_actuator.__name__}?"
    #                 )
    #             agent_actuators.extend(actuator_classes["agent"])
    #             avatar_actuators.extend(actuator_classes["avatar"])

    #     if agent_actuators:
    #         LOGGER.debug(
    #             LOGGER.format_iterable(agent_actuators, message="agent:", indent=True)
    #         )
    #     if avatar_actuators:
    #         LOGGER.debug(
    #             LOGGER.format_iterable(avatar_actuators, message="avatar:", indent=True)
    #         )

    #     return agent_actuators, avatar_actuators

    # def register_task(
    #     self,
    #     name: str,
    #     path: str | List[str],
    #     agent_actuators: List[Callable[[], Actuator]] = None,
    #     avatar_actuators: List[Callable[[], Actuator]] = None,
    #     enable_dynamic_loading: bool = False,
    # ) -> _Task:
    #     """Registers the given task. Loads relevant configuration files.

    #     Args:
    #         name (str): name of the task.
    #         path (str | List[str]): path (or paths) for the task configuration. If multiple paths are specified then files at the earlier paths will be prefered. This allows customisation and overriding of certain configuration files.
    #         agent_actuators (List[Callable[[], Actuator]], optional): classes to be used for agent actuators. Defaults to None.
    #         avatar_actuators (List[Callable[[], Actuator]], optional): classes to be used for avatar actuators. Defaults to None.
    #         enable_dynamic_loading (bool, optional): whether to dynamically load actuators from any `.py` files found in the task path(s). Defaults to False.

    #     Returns:
    #         Task: the configured task
    #     """
    #     if not isinstance(name, str):
    #         raise TypeError("Invalid argument `name`: must be of type `str`")
    #     if isinstance(path, str):
    #         path = [path]
    #     elif isinstance(path, Path):
    #         path = [path.as_posix()]
    #     elif isinstance(path, list):
    #         path = [Path(p).expanduser().resolve().absolute().as_posix() for p in path]
    #     else:
    #         raise TypeError(
    #             "Invalid argument `path`: must be of type `str` or `List[str]`"
    #         )
    #     LOGGER.debug("Registering task: `%s` at path(s): `%s`", name, [p for p in path])
    #     self._template_loader.add_namespace(name, path)
    #     # FILES: <TASK>.svg(.jinja), <TASK>.schema.json AND/OR <TASK>.json, <TASK>.sch
    #     files = self._get_task_files(name)
    #     # print(files)
    #     state = self.load_state(name, files)

    #     agent_actuators = (
    #         copy.deepcopy(agent_actuators) if agent_actuators is not None else []
    #     )
    #     avatar_actuators = (
    #         copy.deepcopy(avatar_actuators) if avatar_actuators is not None else []
    #     )

    #     # *.py (for dynamic loading)
    #     # TODO this should work the same as with config files - i.e. overriding the .py files.
    #     # Currently just loads .py files from all paths, overriding is not yet supported!
    #     # both sets of actuators are updated in the method call
    #     agent_actuators, avatar_actuators = self.load_actuators(
    #         name,
    #         path,
    #         agent_actuators,
    #         avatar_actuators,
    #         enable_dynamic_loading=enable_dynamic_loading,
    #     )

    #     agent_factory = self.load_schedule(
    #         name,
    #         files,
    #         agent_actuators,
    #     )
    #     return _Task(state, agent_factory, avatar_actuators)


def dont_write_bytecode(fun):
    # prevent __pycache__ from being created when dynamically importing
    @wraps(fun)
    def _dont(*args, **kwargs):
        old_dont_write = sys.dont_write_bytecode
        sys.dont_write_bytecode = True
        result = fun(*args, **kwargs)
        sys.dont_write_bytecode = old_dont_write
        return result

    return _dont


@dont_write_bytecode
def load_task_package_from_path(
    path: str | Path, task_name: str, suppress_warnings: bool = False
):
    path = Path(path).expanduser().resolve()
    module_name = f"_{task_name}"
    files = [
        Path(path).expanduser().resolve()
        for path in path.iterdir()
        if path.suffix == ".py"
    ]
    if not files:
        raise TaskConfigurationError(
            f"Failed to load task: `{task_name}`, no `.py` files were found in the task path: `{path}`"
        )
    if "__init__.py" in [path.name for path in files]:
        # load as a package...
        raise NotImplementedError("TODO allow loading task as a package")
    else:
        # load modules individually
        classes = []
        for file in files:
            LOGGER.debug(f"loading task plugin: `{file.name}`")
            classes.extend(_get_classes_from_file(file, f"{module_name}"))
        actuator_classes = _get_actuator_classes(
            classes, suppress_warnings=suppress_warnings
        )
        return actuator_classes, classes


def _get_actuator_classes(
    classes: list[type[Actuator]], suppress_warnings: bool = False
):
    actuators = {"avatar": [], "agent": []}
    for cls in classes:
        if issubclass(cls, Actuator):
            if getattr(cls, CLSATTR_IS_AGENT_ACTUATOR, False):
                actuators["agent"].append(cls)
            elif getattr(cls, CLSATTR_IS_AVATAR_ACTUATOR, False):
                actuators["avatar"].append(cls)
            elif not suppress_warnings:
                LOGGER.warning(
                    "An Actuator definition was found: `%s`, but it was not tagged as @%s or @%s.",
                    cls.__name__,
                    agent_actuator.__name__,
                    avatar_actuator.__name__,
                )
    return actuators


def _get_classes_from_file(file: str | Path, module_name: str):
    module = _get_module_from_file(file, module_name)
    classes = [
        member
        for _, member in inspect.getmembers(module)
        if inspect.isclass(member) and member.__module__ == module_name
    ]
    return classes


def _get_module_from_file(file: str | Path, module_name: str):
    spec = importlib.util.spec_from_file_location(module_name, str(file))
    if spec is not None:
        module = importlib.util.module_from_spec(spec)
        # sys.modules[module_name] = module  # Optional: Add to sys.modules if you want it globally available
        spec.loader.exec_module(module)
        return module
    else:
        raise ImportError(f"Failed to load task module from the given path: `{file}`")



def bounding_rectangle(rectangles):
    """Computes the bounding rectangle given a list of rectangles represented as dictionaries.

    Args:
        rectangles (list[dict]): A list of dictionaries representing rectangles.
            Each dictionary should have keys 'x', 'y', 'width', and 'height' representing
            the x and y coordinates of the top-left corner, and the width and height of
            each rectangle, respectively.

    Returns:
        dict: A dictionary representing the bounding rectangle.
            The dictionary contains keys 'x', 'y', 'width', and 'height' representing
            the x and y coordinates of the top-left corner, and the width and height of
            the bounding rectangle, respectively.

    Example:
        >>> rectangles = [
        ...     {"x": 0, "y": 0, "width": 320, "height": 320},
        ...     {"x": 100, "y": 100, "width": 200, "height": 200},
        ...     {"x": 150, "y": 150, "width": 50, "height": 50},
        ... ]
        >>> compute_bounding_rectangle(rectangles)
        {'x': 0, 'y': 0, 'width': 420, 'height': 420}
    """
    # Initialize min and max values
    min_x = float("inf")
    min_y = float("inf")
    max_x = float("-inf")
    max_y = float("-inf")

    # Iterate over the rectangles to find min and max values
    for rect in rectangles:
        min_x = min(min_x, rect["x"])
        min_y = min(min_y, rect["y"])
        max_x = max(max_x, rect["x"] + rect["width"])
        max_y = max(max_y, rect["y"] + rect["height"])

    # Compute width and height of bounding rectangle
    width = max_x - min_x
    height = max_y - min_y

    # Construct the bounding rectangle
    bounding_rect = {"x": min_x, "y": min_y, "width": width, "height": height}

    return bounding_rect



"""Module containing the `AvatarActuator`, see class documentation for details."""

from star_ray.agent import Actuator, attempt

from icua.event import RenderEvent, UserInputEvent


class AvatarActuator(Actuator):
    """This actuator will forward all user generated events to the environment. Other agents may subscribe to receive these events, otherwise they will simply be logged. It may also be used to produce `RenderEvent`s which should be taken at the START of the UI rendering step performed by the avatar."""

    @attempt
    def render(self, action: RenderEvent = None):
        """Attempt method that will attempt a `RenderEvent`."""
        if action is None:
            return RenderEvent()
        return action

    @attempt
    def default(self, action: UserInputEvent):
        """Default attempt method for this actuator, will simply forward any of the received types to the environment. Typically it will recieve all types present in `UserInputEvent` (assuming that the actuator is attached to an `Avatar`.

        Args:
            action (UserInputEvent): user input event.

        Returns:
            Event: the given `action`
        """
        return action



"""Module defines an actuator that will take actions indicating that a task has changed its acceptability status (see `icua.event.TaskAcceptable` and `icua.event.TaskUnacceptable`)."""

from star_ray import Actuator, attempt
from icua.event import TaskAcceptable, TaskUnacceptable


class TaskAcceptabilityActuator(Actuator):
    """Abstract base class for guidance actuators."""

    @attempt
    def on_acceptable(self, action: TaskAcceptable):
        """An attempt method that will show guidance to a user."""
        return action

    @attempt
    def on_unacceptable(self, action: TaskUnacceptable):
        """An attempt method that will hide guidance from a user."""
        return action



"""Package defines or includes many useful event classes including task, guidance, UI, user input and XML/SVG related events."""

from star_ray_pygame.event import (
    MouseButtonEvent,
    MouseMotionEvent,
    KeyEvent,
    WindowCloseEvent,
    WindowOpenEvent,
    WindowFocusEvent,
    WindowMoveEvent,
    WindowResizeEvent,
    ScreenSizeEvent,
)

from star_ray_xml import (
    XMLQuery,
    XMLUpdateQuery,
    XPathQuery,
    Select,
    Update,
    Replace,
    Insert,
    Delete,
)

from .event_task import EnableTask, DisableTask

from .event_guidance import (
    TaskAcceptable,
    TaskUnacceptable,
    ShowGuidance,
    HideGuidance,
    XPathAction,
    DrawArrowAction,
    DrawBoxAction,
    DrawBoxOnElementAction,
    ShowElementAction,
    HideElementAction,
)

from .event_avatar import RenderEvent

# TODO try-except this?
from ..extras.eyetracking import EyeMotionEvent, EyeMotionEventRaw

USER_INPUT_TYPES: tuple[type] = (
    EyeMotionEvent,
    EyeMotionEventRaw,
    MouseMotionEvent,
    MouseButtonEvent,
    KeyEvent,
    WindowOpenEvent,
    WindowCloseEvent,
    WindowFocusEvent,
    WindowMoveEvent,
    WindowResizeEvent,
    ScreenSizeEvent,
)

UserInputEvent = (
    MouseMotionEvent
    | MouseButtonEvent
    | KeyEvent
    | WindowOpenEvent
    | WindowCloseEvent
    | WindowFocusEvent
    | WindowMoveEvent
    | WindowResizeEvent
    | ScreenSizeEvent
    | EyeMotionEvent
    | EyeMotionEventRaw
)


__all__ = (
    # user input events
    "UserInputEvent",
    "MouseEvent",
    "WindowEvent",
    "EyeMotionEvent",
    "EyeMotionEventRaw",
    "MouseButtonEvent",
    "MouseMotionEvent",
    "KeyEvent",
    "WindowCloseEvent",
    "WindowFocusEvent",
    "WindowMoveEvent",
    "WindowOpenEvent",
    "WindowResizeEvent",
    "ScreenSizeEvent",
    # xml/svg events
    "XMLQuery",
    "XMLUpdateQuery",
    "XPathQuery",
    "Update",
    "Select",
    "Insert",
    "Delete",
    "Replace",
    # task events
    "EnableTask",
    "DisableTask",
    "TaskAcceptable",
    "TaskUnacceptable",
    # guidance events
    "ShowGuidance",
    "HideGuidance",
    "XPathAction",
    "DrawArrowAction",
    "DrawBoxAction",
    "DrawBoxOnElementAction",
    "ShowElementAction",
    "HideElementAction",
    # misc
    "RenderEvent",
)



import sys
import aiostream
import asyncio
import itertools
import inspect
from typing import Any
from collections.abc import Callable
from star_ray import Agent, Actuator
from star_ray.event import ErrorObservation
from pyfuncschedule import parser as schedule_parser, Schedule, ScheduleParser
from ._logging import LOGGER
from ._error import TaskConfigurationError


class ScheduledAgent2(Agent):
    def __init__(self, actuators: list[Actuator], schedules: list[Schedule]):
        pass


class ScheduledAgent(Agent):
    """The actions of this agent are determined by a fixed schedule that is provided upon creation. The schedule is a collection of `attempts` that are part of the given actuators."""

    def __init__(self, actuators: list[Actuator], schedules: list[Schedule]):
        super().__init__([], actuators)
        self._schedules = [sch.stream() for sch in schedules]
        self._iter_context = None  # set on initialise
        self._iter_schedules = None  # set on initialise
        self._completed = False

    async def __initialise__(self, state):
        self._iter_context = aiostream.stream.merge(*self._schedules).stream()
        self._iter_schedules = await self._iter_context.__aenter__()
        LOGGER.debug(f"{ScheduledAgent.__name__}: {self.id} initialised")

    # TODO implement proper execution scheduling in the environment!! we cant have each agent awaiting here blocking each other
    async def __cycle__(self):
        # check if there were any errors from actuators
        for actuator in self.actuators:
            for obs in actuator.iter_observations():
                if isinstance(obs, ErrorObservation):
                    raise obs.exception()
        if self._completed:
            await asyncio.sleep(0.1)  # TODO this is a temporary
            return  # TODO the agent is done, the environment should not be calling __cycle__ for this agent?
        try:
            # this will await the next action from the collection of schedules
            await self._iter_schedules.__anext__()
            # print(len(self._schedules), action)
        except StopAsyncIteration:
            await self._iter_context.__aexit__(None, None, None)
            self._completed = True
            LOGGER.debug(f"{self} has completed all its scheduled events.")
        except asyncio.CancelledError:
            await self._iter_context.__aexit__(None, None, None)
            self._completed = True
            LOGGER.debug(f"{self} was cancelled.")
        except Exception as e:
            exc_type, exc_val, exc_tb = sys.exc_info()
            await self._iter_context.__aexit__(exc_type, exc_val, exc_tb)
            raise e

    async def __sense__(self, state, *args, **kwargs):
        return super().__sense__(state, *args, **kwargs)

    async def __execute__(self, state, *args, **kwargs):
        return super().__execute__(state, *args, **kwargs)

    async def __terminate__(self):
        await self._iter_context.aclose()


class ScheduledAgentFactory:
    def __init__(
        self,
        schedule_source: str,
        actuator_types: list[type[Actuator]],
        funcs: list[Callable],
    ):
        super().__init__()
        self._source = schedule_source
        self._actuator_types = list(set(actuator_types))
        self._functions: dict[str, Callable] = {fun.__name__: fun for fun in funcs}
        self._parse_result: Any = None  # TODO type hint...
        self.parse_schedule()

    def parse_schedule(
        self,
    ):
        parser = ScheduleParser()
        for name, fun in self._functions.items():
            parser.register_function(fun, name=name)
            LOGGER.debug(f"registered function: {name}")

        self._attempts = dict()
        for cls, action in self._get_all_attempt_methods(self._actuator_types):
            parser.register_action(action)
            LOGGER.debug(f"registered attempt: {cls.__name__}@{action.__name__}")
        try:
            self._parse_result = parser.parse(self._source)
        except Exception as e:
            raise TaskConfigurationError("Failed to parse schedule.") from e

        try:
            # this will not resolve fully because we are using the unbound attempt methods
            # this is just to validate the attempt methods/functions being used after parsing
            _ = parser.resolve(self._parse_result)
        except Exception as e:
            raise TaskConfigurationError("Failed to validate schedule.") from e

    def __call__(self) -> ScheduledAgent:
        actuators = [cls() for cls in self._actuator_types]
        attempts = {
            attempt.__name__: attempt
            for _, attempt in self._get_all_attempt_methods(actuators)
        }
        # create schedule from bound methods. these methods are now tied to the list of actuators!
        schedules = schedule_parser.resolve(
            self._parse_result, attempts, self._functions
        )
        return ScheduledAgent(actuators, schedules)

    def _iter_attempt_methods_unbound(self, actuator: type[Actuator]):
        assert issubclass(actuator, Actuator)
        methods = inspect.getmembers(actuator, predicate=inspect.isfunction)
        methods = [m[1] for m in filter(lambda m: hasattr(m[1], "is_attempt"), methods)]
        for fun in methods:
            yield actuator, fun

    def _iter_attempt_methods_bound(self, actuator: Actuator):
        assert isinstance(actuator, Actuator)
        methods = inspect.getmembers(actuator, predicate=inspect.ismethod)
        methods = [m[1] for m in filter(lambda m: hasattr(m[1], "is_attempt"), methods)]
        for fun in methods:
            yield actuator, fun  # bind

    def _get_all_attempt_methods(self, actuators: list[Actuator | type[Actuator]]):
        if len(actuators) == 0:
            return iter([])
        if isinstance(actuators[0], Actuator):
            return itertools.chain(
                *[self._iter_attempt_methods_bound(actu) for actu in actuators]
            )
        elif issubclass(actuators[0], Actuator):
            return itertools.chain(
                *[self._iter_attempt_methods_unbound(actu) for actu in actuators]
            )
        else:
            raise TypeError(f"Invalid type: {type(actuators[0])}")



"""Module contains eyetracking event classes: `EyeMotionEvent` and `EyeMotionEventRaw`, see class documentation for details."""

from pydantic import Field
from star_ray.event import Event

__all__ = ("EyeMotionEvent", "EyeMotionEventRaw")


class EyeMotionEventRaw(Event):
    """A class representing an eyetracking motion event.

    Attributes:
        position (tuple[float,float]): The position of the eyes relative to the physical monitor or screen (typically in normalised range [0,1]).
    """

    position: tuple[float, float] | tuple[int, int]


class EyeMotionEvent(Event):
    """A class representing an eye tracking motion event with additional UI window information.

    Attributes:
        position (tuple[float,float]): The position of the eyes relative to the UI in svg space.
        position_raw  (tuple[float,float]): The position of the eyes relative to the UI in window space (pixels).
        position_screen (tuple[float,float]): The position of the eyes relative to the physical monitor or screen (typically in normalised range [0,1]).
        in_window (bool): Whether the eyes are within the ui window.
        target (list[str]): The UI elements that the eyes are currently over. This value is UI implementation dependent and may be None, typically it will contain unique element ids.
        fixated: (bool): Whether the eye are fixated or saccading based on gaze velocity or acceleration.
    """

    position: tuple[float, float] | tuple[int, int]
    position_raw: tuple[float, float] | tuple[int, int]
    position_screen: tuple[float, float] | tuple[int, int] | None
    fixated: bool
    in_window: bool
    target: list[str] | None = Field(default_factory=lambda: None)



"""Module containing useful tools for plotting event occurances."""

from typing import Literal
import matplotlib.pyplot as plt
import numpy as np
from star_ray import Event


def _get_fig_ax(ax: plt.Axes | None = None):
    if ax is None:
        fig, ax = plt.subplots(figsize=(12, 2))
        ax.set_yticks([])
        fig.tight_layout()
    else:
        fig = ax.get_figure()
    return fig, ax


def plot_intervals(
    timestamps: np.ndarray,
    color: str = "blue",
    alpha: float = 1.0,
    # linestyle: str = "-",
    # linewidth: float = 1.0,
    ymin: float = 0.0,
    ymax: float = 1.0,
    ax: plt.Axes | None = None,
):
    """TODO."""
    assert len(timestamps.shape) == 2
    assert timestamps.shape[1] == 2
    fig, ax = _get_fig_ax(ax)
    for interval in timestamps:
        ax.axvspan(
            interval[0],
            interval[1],
            color=color,
            alpha=alpha,
            ymin=ymin,
            ymax=ymax,
            # linestyle=linestyle,
            # linewidth=linewidth,
        )
    (xmin, xmax) = ax.get_xlim()
    ax.set_xlim(min(xmin, timestamps.min()), max(xmax, timestamps.max()))
    return fig


def plot_timestamps(
    events: list[tuple[float, Event]],
    cls: type,
    mode: Literal["log", "event"] = "log",
    color: str = "blue",
    alpha: float = 1.0,
    linestyle: str = "-",
    linewidth: float = 1.0,
    ymin: float = 0.0,
    ymax: float = 1.0,
    ax: plt.Axes | None = None,
):
    """TODO."""
    events = filter(lambda x: isinstance(x[1], cls), events)

    if mode == "log":
        ts = np.array([x[0] for x in events])
    elif mode == "event":
        ts = np.array([x[1].timestamp for x in events])
    else:
        raise TypeError(f"Invalid mode: {mode}, must be one of {mode.__args__}")

    fig, ax = _get_fig_ax(ax)
    for t in ts:
        ax.axvline(
            t,
            ymin=ymin,
            ymax=ymax,
            color=color,
            linestyle=linestyle,
            linewidth=linewidth,
            alpha=alpha,
        )
    return fig



"""Package contains environment related functionality, including support for "multi-task" environments.

See:
    - `MultiTaskEnvironment` and `MultiTaskAmbient` classes for multi-task support.
"""

from .multitask_ambient import MultiTaskAmbient
from .multitask_environment import MultiTaskEnvironment

__all__ = (
    "MultiTaskEnvironment",
    "MultiTaskAmbient",
)



"""Module defines the `Avatar` class, which will display a UI (`pygame` backend) and capture user input, including mouse clicks, motion, key events, window events, and eyetracking events (if the required sensor is added, see `EyetrackingIOSensor`)."""

from star_ray import Sensor, Actuator, observe
from star_ray_pygame.avatar import Avatar as PygameAvatar
from star_ray_pygame import WindowConfiguration
from icua.event import (
    RenderEvent,
    EyeMotionEvent,
)


class Avatar(PygameAvatar):
    """A special type of `Agent` that will display a UI (`pygame` backend) to the user and capture various kinds of user input, including: mouse clicks, motion, key events, window events, and eyetracking events (if the required sensor is added, see `EyetrackingIOSensor`)."""

    def __init__(
        self,
        sensors: list[Sensor],
        actuators: list[Actuator],
        window_config: WindowConfiguration = None,
        **kwargs,
    ):
        """Constructor.

        Args:
            sensors (list[Sensor]): list of initial sensors, this may include instances of `IOSensor` that will gather user input from devices (see e.g. `EyetrackingIOSensor`). Common user input events such as mouse, keyboard and window events are already handled by a io sensor that is added automatically.
            actuators (list[Actuator]): list of initial actuators, this should generally include an actuator that is capable of forwarding user input events (see e.g. `AvatarActuator`).
            window_config (WindowConfiguration, optional): UI window configuration. Defaults to None (see `star_ray_pygame.agent.PyGameAvatar` for details).
            kwargs (dict[str,Any]): additional optional keyword arguments.
        """
        super().__init__(sensors, actuators, window_config=window_config, **kwargs)

    def render(self) -> None:
        """Renders the UI and triggers a `RenderEvent`."""
        # this is for logging purposes, we can see when the rendering beings. if all agents are running locally (and synchronously), then we can assume that all preceeding events in the event log will be visible to the user! this is very useful for post-analysis in experiments.
        self.attempt(RenderEvent())
        super().render()

    @observe
    def on_gaze(self, event: EyeMotionEvent):
        """Observe method for `EyeMotionEvents`, this will only be called if an `EyetrackingIOSensor` (or similar) is attached to this agent. It finalises the events by computing the svg coordinates (from window coordinates) and finding all svg elements that lie under this coordinate. It will attempt the event so that it is made avaliable (via an appropriate actuator) to other subscribing agents.

        Args:
            event (EyeMotionEvent): the eye motion event.
        """
        # These events are generated by an EyetrackingIOSensor (if it exists).
        # The position is in pixel-coordinates, we need to transform to SVG coordinates
        # (similar to mouse events).
        event.position = self._view.pixel_to_svg(event.position_raw)
        # also find the elements that are under the gaze point
        event.target = self._view.elements_under(event.position, transform=False)
        # attempt the event, this will send it to other subscribing agents
        self.attempt(event)



"""Module containing events that may be useful for an avatar or for event logging."""

from star_ray.event import Event


class RenderEvent(Event):
    """Event used to indicate that the UI is being rendered by an avatar, has no effect and is used for logging purposes."""



"""Defines a `Sensor` that will subscribe to all user input events by default: `EyeMotionEvent`, `MouseMotionEvent`, `MouseButtonEvent`, `KeyEvent`, `WindowOpenEvent`, `WindowCloseEvent`, `WindowFocusEvent`, `WindowMoveEvent`, `WindowResizeEvent`."""

from typing import Any
from star_ray import Event, Sensor
from star_ray.pubsub import Subscribe

from ..event import USER_INPUT_TYPES


class UserInputSensor(Sensor):
    """Sensor that will subscribe to user input events."""

    def __init__(
        self,
        subscribe_to: tuple[type[Event]] = None,
        *args: list[Any],
        **kwargs: dict[str, Any],
    ):
        """Constructor.

        This sensor will subscribe to the following events: `EyeMotionEvent`, `MouseMotionEvent`, `MouseButtonEvent`, `KeyEvent`, `WindowOpenEvent`, `WindowCloseEvent`, `WindowFocusEvent`, `WindowMoveEvent`, `WindowResizeEvent`.

        Additional events can be added via the `subscribe_to` argument.

        Args:
            subscribe_to (tuple[Type[Event]], optional): additional user input event types to subscribe to. Defaults to None.
            args (list[Any]): Additional optional arguments.
            kwargs (dict[str,Any]): Additional optionals keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._subscribe_to = list(USER_INPUT_TYPES)
        self._subscribe_to.extend(subscribe_to if subscribe_to else [])
        # remove duplicates if there are any
        self._subscribe_to = list(set(self._subscribe_to))

    def __subscribe__(self):  # noqa
        return [Subscribe(topic=event_type) for event_type in self._subscribe_to]



"""Task acceptability functionality. The `TaskAcceptabilitySensor` can be extended to track the state of a task when used as part of a `icua.agent.GuidanceAgent`."""

from abc import abstractmethod
from typing import Any
from pydantic import computed_field
from copy import deepcopy
from star_ray.agent import Sensor
from star_ray.event import Event, Observation, ErrorObservation
from star_ray_xml import Select

__all__ = ("TaskAcceptabilityObservation", "TaskAcceptabilitySensor")


class TaskAcceptabilityObservation(Observation):
    """Observation representing the acceptability and activity of a given task.

    The `values` attribute contains the following fields:
    - `task` (str): the name of the task.
    - `is_active` (bool): whether the task is currently active.
    - `is_acceptable` (bool): whether the task is in an acceptable state, this is always False if `is_active` is False.
    """

    @computed_field
    @property
    def task(self) -> str:
        """Getter for the name of the task."""
        return self.values["task"]

    @computed_field
    @property
    def is_active(self) -> str:
        """Whether the task is currently active."""
        return self.values["is_active"]

    @computed_field
    @property
    def is_acceptable(self) -> str:
        """Whether the task is in an acceptable state, this is always False if `is_active` is False."""
        return self.values["is_acceptable"]


class TaskAcceptabilitySensor(Sensor):
    """This `Sensor` can be used by an agent to track the acceptability of a task."""

    def __init__(self, task_name: str, *args: list[Any], **kwargs: dict[str, Any]):
        """Constructor.

        Args:
            task_name (str): task to track.
            args (list[Any]): Additional optional arguments.
            kwargs (dict[str,Any]): Additional optionals keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self._task_name = task_name
        self._beliefs = dict()
        self._errors = []

    @abstractmethod
    def is_acceptable(self, task: str = None, **kwargs) -> bool:
        """Checks whether the task(s) tracked by this `Sensor` is in an acceptable state given current beliefs about it.

        Args:
            task (str, optional): task to check (if this `Sensor` is tracking more than one task). Defaults to None, which indicate the main task this `Sensor` is tracking.
            kwargs (Dict[str, Any], optional): any additional arguments
        Returns:
            bool: whether the task is acceptable.
        """

    @abstractmethod
    def is_active(self, task: str = None, **kwargs) -> bool:
        """Checks whether the task(s) tracked by this `Sensor` is active in the environment.

        Args:
            task (str, optional): task to check (if this `Sensor` is tracking more than one task). Defaults to None, which indicate the main task this `Sensor` is tracking.
            kwargs (Dict[str, Any], optional): any additional arguments
        Returns:
            bool: whether the task is active.
        """

    @abstractmethod
    def sense(self) -> list[Select]:
        """Sense actions that are relevant for tracking the acceptability for the given task. This method will be called automatically during the sense cycle of this `Sensor`. The resulting observations are stored internally and used to determine the acceptability of the given task.

        Returns:
            List[Select]: sense actions.
        """

    def sense_element(
        self, element_id: str = None, xpath: str = None, attributes: list[str] = None
    ) -> Select:
        """Factory for a sensor action that will sense an element with a given `id`.

        Args:
            element_id (str, optional): `id` of the element to sense  (if `xpath` is not provided). Defaults to None.
            xpath (str, optional): xpath of the element (if `element_id` is not provided). Defaults to None.
            attributes (list[str], optional): element attributes to sense. Defaults to None.

        Raises:
            ValueError: if both `element_id` and `xpath` are provided.

        Returns:
            Select: the sense action.
        """
        # TODO maybe this could be in a parent class? e.g. an XMLSensor?
        if element_id is None and xpath is None:
            raise ValueError("One of: `element_id` or `xpath` must be specified.")
        elif element_id:
            xpath = f"//*[@id='{element_id}']"
        if attributes is None:
            attributes = []
        if "id" not in attributes:
            attributes.append("id")
        return Select(source=self.id, xpath=xpath, attrs=attributes)

    def iter_observations(self):  # noqa
        yield from self._errors
        self._errors.clear()
        is_active = self.is_active(self.task_name)
        if not is_active:
            yield TaskAcceptabilityObservation(
                source=self.id,
                values=dict(
                    task=self.task_name, is_active=is_active, is_acceptable=False
                ),
            )
        else:
            yield TaskAcceptabilityObservation(
                source=self.id,
                values=dict(
                    task=self.task_name,
                    is_active=is_active,
                    is_acceptable=self.is_acceptable(),
                ),
            )

    def __sense__(self) -> list[Event]:  # noqa
        actions = self.sense()
        # sense() must return a list of actions!
        if not isinstance(actions, list | tuple):
            raise TypeError(
                f"`sense()` must return a `list` of events, received: {type(actions)}"
            )
        return actions

    def __transduce__(self, observations: list[Observation]) -> list[Observation]:  # noqa
        # This is called whenever a new observation is sensed by this sensor (either via __query__ or __notify__).
        # The observations are consumed by the sensor to update the current beliefs.
        # New observations will be generated in `iter_observations` based on the beliefs,
        # typically these will indicate if the acceptability of the task has changed.
        # The agent can also access this information directly using the `is_acceptable` method.
        for observation in observations:
            self._update_beliefs(observation)
        return []

    def on_error_observation(self, observation: ErrorObservation):
        """Handle observation errors, which may occur if for example, required task elements are missing during sense actions. By default these error observations will be produced by `iter_observations` delegating the error handling to the agent. Override this method if you want custom behaviour for handling errors.

        Args:
            observation (ErrorObservation): error observation.
        """
        self._errors.append(observation)

    @property
    def beliefs(self) -> dict[str, Any]:
        """The `Sensor's` current beliefs about the state of the task.

        Returns:
            Dict[str, Any]: beliefs about the state of the task.
        """
        return self._beliefs

    @property
    def task_name(self) -> str:
        """The name of the task that this `Sensor` is tracking.

        Returns:
            _type_: _description_
        """
        return self._task_name

    def _update_beliefs(self, observation: Observation | ErrorObservation):
        if isinstance(observation, ErrorObservation):
            self.on_error_observation(observation)
        elif isinstance(observation, Observation):
            # TODO observation.values should only be of length 1?
            for data in observation.values:
                try:
                    self._beliefs[data.pop("id")] = deepcopy(data)
                except KeyError:
                    raise KeyError(
                        f"Observation: {observation} doesn't contain the required `id` attribute."
                    )
        else:
            raise TypeError(f"Invalid observation type: {type(observation)}")



"""Defines the `ScheduledAgent` class. Agents of this type work with `pyfuncschedule` to periodically execute actions (according to the schedule)."""

import asyncio
import itertools
import inspect

from typing import Generic, TypeVar

from collections.abc import Callable, Iterator, AsyncGenerator
from star_ray import Actuator
import bisect
import time

from star_ray import Agent
from star_ray.environment import State
from star_ray.event import ErrorObservation
from pyfuncschedule import parser as schedule_parser, Schedule, ScheduleParser
from ._logging import LOGGER
from ._error import TaskConfigurationError


class ScheduledAgent(Agent):
    """The actions of this agent are determined by a fixed schedule that is provided upon creation. The schedule is a collection of `attempts` that are part of the given actuators."""

    def __init__(self, actuators: list[Actuator], schedules: list[Schedule]):
        super().__init__([], actuators)
        self._schedules = [iter(sch) for sch in schedules]
        self._next_items = []
        t = time.time()
        # Load the first item from each schedule iterator
        for it in self._schedules:
            try:
                (dt, value) = next(it)
                bisect.insort(
                    self._next_items, ((t + dt, value), it), key=lambda x: x[0][0]
                )
            except StopIteration:
                continue  # Skip if iterator is initially empty

    @property
    def _completed(self):
        return len(self._next_items) == 0

    def __cycle__(self):
        # print(self._next_items)
        # check if there were any errors from actuators
        for actuator in self.actuators:
            for obs in actuator.iter_observations():
                if isinstance(obs, ErrorObservation):
                    raise obs.exception()

        if self._completed:
            LOGGER.warning(
                f"{self} `__cycle__` was called after all schedules have completed."
            )

        while not self._completed:
            # get the earliest item
            (et, attempt), it = self._next_items[0]
            t = time.time()
            nt = et - t
            if nt <= 0:
                # execute immediately and continue to next soonest item
                self._next_items = self._next_items[1:]
                try:
                    (dt, value) = next(it)
                    # nt will compensate for any overshooting
                    bisect.insort(
                        self._next_items,
                        ((t + dt + nt, value), it),
                        key=lambda x: x[0][0],
                    )
                except StopIteration:
                    pass  # no more values, the schedule is complete, try next schedule
                attempt()  # attempt the action
            else:
                break  # no more events scheduled for this cycle


class ScheduledAgentAsync(Agent):
    """The actions of this agent are determined by a fixed schedule that is provided upon creation. The schedule is a collection of `attempts` that are part of the given actuators."""

    def __init__(self, actuators: list[Actuator], schedules: list[Schedule]):
        super().__init__([], actuators)
        self._schedules = [iter(sch) for sch in schedules]
        self._schedules_iter = merge_iterators(self._schedules)
        self._completed = False

    async def __sense__(self, state, *args, **kwargs):
        return super().__sense__(state, *args, **kwargs)

    async def __execute__(self, state, *args, **kwargs):
        return super().__execute__(state, *args, **kwargs)

    async def __terminate__(self):
        self._completed = True
        await self._schedules_iter.aclose()

    async def __initialise__(self, state: State):
        return super().__initialise__(state)

    async def __cycle__(self):
        # check if there were any errors from actuators
        for actuator in self.actuators:
            for obs in actuator.iter_observations():
                if isinstance(obs, ErrorObservation):
                    raise obs.exception()

        if self._completed:
            LOGGER.warning(
                f"{self} `__cycle__` was called after all schedules have completed."
            )
            return  # TODO the agent should take an action which terminates itself!
        try:
            # this will await the next action from the collection of schedules
            attempt = await self._schedules_iter.__anext__()
            attempt()  # attempt the action
            # repeatedly try and get elements if they are yielded in a very short window,
            # if a timeout occurs then yield control back to the main event loop
            # don't try take too many elements, otherwise this may hog the event loop
            # for _ in range(1000):
            #     try:
            #         print("__")
            #         attempt = await asyncio.wait_for(
            #             self._schedules_iter.__anext__(), 0
            #         )
            #         attempt()  # attempt the action
            #     except asyncio.TimeoutError:
            #         print("?")
            #         break  # no more items were immediately avaliable
        except StopAsyncIteration:
            self._completed = True
            LOGGER.debug(f"{self} has completed all its scheduled events.")
        except asyncio.CancelledError:
            self._completed = True
            LOGGER.debug(f"{self} was cancelled.")


class ScheduledAgentFactory:
    def __init__(
        self,
        schedule_source: str,
        actuator_types: list[type[Actuator]],
        funcs: list[Callable],
    ):
        super().__init__()
        self._source = schedule_source
        self._actuator_types = list(set(actuator_types))
        self._functions: dict[str, Callable] = {fun.__name__: fun for fun in funcs}
        self._parse_result: list[Schedule] = None
        self.parse_schedule()

    def parse_schedule(
        self,
    ):
        parser = ScheduleParser()
        for name, fun in self._functions.items():
            parser.register_function(fun, name=name)
            LOGGER.debug(f"registered schedule function: {name}")

        self._attempts = dict()
        for cls, action in self._get_all_attempt_methods(self._actuator_types):
            parser.register_action(action)
            LOGGER.debug(
                f"registered schedule attempt: {cls.__name__}@{action.__name__}"
            )
        try:
            self._parse_result = parser.parse(self._source)
        except Exception as e:
            raise TaskConfigurationError("Failed to parse schedule.") from e

        try:
            # this will not resolve fully because we are using the unbound attempt methods
            # this is just to validate the attempt methods/functions being used after parsing
            _ = parser.resolve(self._parse_result)
        except Exception as e:
            raise TaskConfigurationError("Failed to validate schedule.") from e

    def __call__(self) -> ScheduledAgent:
        actuators = [cls() for cls in self._actuator_types]
        attempts = {
            attempt.__name__: attempt
            for _, attempt in self._get_all_attempt_methods(actuators)
        }
        # create schedule from bound methods, these methods are now tied to the actuators!
        schedules = schedule_parser.resolve(
            self._parse_result, attempts, self._functions
        )
        return ScheduledAgent(actuators, schedules)

    def _iter_attempt_methods_unbound(self, actuator: type[Actuator]):
        assert issubclass(actuator, Actuator)
        methods = inspect.getmembers(actuator, predicate=inspect.isfunction)
        methods = [m[1] for m in filter(lambda m: hasattr(m[1], "is_attempt"), methods)]
        for fun in methods:
            yield actuator, fun

    def _iter_attempt_methods_bound(self, actuator: Actuator):
        assert isinstance(actuator, Actuator)
        methods = inspect.getmembers(actuator, predicate=inspect.ismethod)
        methods = [m[1] for m in filter(lambda m: hasattr(m[1], "is_attempt"), methods)]
        for fun in methods:
            yield actuator, fun

    def _get_all_attempt_methods(self, actuators: list[Actuator | type[Actuator]]):
        if len(actuators) == 0:
            return iter([])
        if isinstance(actuators[0], Actuator):
            return itertools.chain(
                *[self._iter_attempt_methods_bound(actu) for actu in actuators]
            )
        elif issubclass(actuators[0], Actuator):
            return itertools.chain(
                *[self._iter_attempt_methods_unbound(actu) for actu in actuators]
            )
        else:
            raise TypeError(f"Invalid type: {type(actuators[0])}")


V = TypeVar("V")


async def merge_iterators(
    iterators: list[Iterator[tuple[float, Generic[V]]]],
) -> AsyncGenerator[V]:
    """Merge iterators to create an async iterator that waits for a given time. Each iterator should yield (dt, value). The resulting async iterator will wait (approximately) dt time before yielding the iterator value.

    Args:
        iterators (list[Iterator[tuple[float, V]]]): iterators to merge

    Yields:
        V: values yielded by inner iterators
    """
    # Initialize a list to keep track of the next items from each iterator
    next_items = []

    t = time.time()  # start time
    # Load the first item from each iterator and insert it into the sorted list
    for it in iterators:
        try:
            (dt, value) = next(it)
            bisect.insort(next_items, ((t + dt, value), it), key=lambda x: x[0][0])
        except StopIteration:
            continue  # Skip if iterator is initially empty

    # Process items until all iterators are exhausted
    while next_items:
        # get the earliest item
        (et, value), it = next_items[0]
        t = time.time()
        nt = et - t
        if nt <= 0:
            # execute immediately and continue to next soonest item
            next_items = next_items[1:]
            try:
                (dt, next_value) = next(it)
                # nt will compensate for any overshooting
                bisect.insort(
                    next_items, ((t + dt + nt, next_value), it), key=lambda x: x[0][0]
                )
            except StopIteration:
                pass  # no more values to add the iterator is complete
            yield value
            # allow other tasks to run if iterator dts are too fast...
            await asyncio.sleep(0)
        else:
            # wait for some time
            wait = max(nt / 4, 0.001)
            await asyncio.sleep(wait)



"""Package contains eyetracking related functionality. It supports common eyetracker SDKs and integrates them with the `star_ray` event API.

See:
    - `EyetrackerBase`: base class for all eyetrackers.
    - `tobii` package: implementation of `EyetrackerBase` using tobii eyetracking SDK (requires `tobii_research` package).
"""

from .eyetrackerbase import EyetrackerBase
from .event import EyeMotionEvent, EyeMotionEventRaw
from .filter import IVTFilter, NWMAFilter, WindowSpaceFilter
from .sensor import EyetrackerIOSensor

# requires extra "tobii"
from . import tobii

__all__ = (
    # events
    "EyeMotionEvent",
    "EyeMotionEventRaw",
    "EyetrackerIOSensor",
    # filters
    "IVTFilter",
    "NWMAFilter",
    "WindowSpaceFilter",
    # eyetracker base
    "EyetrackerBase",
    "tobii",  # tobii base eyetracker
)



"""Module that contains the abstract class `EyetrackerBase` which provides a straight forward API for eyetracking. Typically an `EyetrackerBase` subclass is implemented via an eyetracker hardware provider SDK (see e.g. `TobiiEyetracker`)."""

from abc import ABC, abstractmethod
from star_ray import Event


class EyetrackerBase(ABC):
    """Abstract base class for eyetrackers."""

    @abstractmethod
    def start(self) -> None:
        """Start the eyetracker."""

    @abstractmethod
    def stop(self) -> None:
        """Stop the eyetracker."""

    @abstractmethod
    async def get(self) -> list[Event]:
        """Get the most recent eyetracking events. Will await the next event if there are none pending.

        Returns:
            list[Event]: recent eyetracking events (most recent first).
        """

    @abstractmethod
    def get_nowait(self) -> list[Event]:
        """Get the most recent eyetracking events, will return an empty list if there are no pending events.

        Returns:
            list[Event]: recent eyetracking events (most recent first).
        """



"""Module that defines the `GuidanceAgent` class. This type of agent can be used to provide visual feedback or "guidance" to a user based on their actions and the current state of the environment, see class documentation for details."""

from typing import Any
from collections.abc import Iterator
from collections import defaultdict, deque
from itertools import islice
from star_ray.agent import AgentRouted, Component, Sensor, Actuator, observe
from star_ray.event import Event, ErrorObservation, ErrorActiveObservation

from .sensor_acceptability import TaskAcceptabilityObservation
from .sensor_userinput import UserInputSensor
from .actuator_acceptability import TaskAcceptabilityActuator
from ..event import TaskAcceptable, TaskUnacceptable


class GuidanceAgent(AgentRouted):
    """Base class for an agent that can provide users with visual guidance. It makes use of the task acceptability sensor API (see `icua.agent.TaskAcceptabilitySensor`) to determine which tasks may require the users attention.

    The following methods will be called at the appropriate moment (e.g. when a task switches from an acceptable state to an unacceptable state), a guidance agent should implement these to decide what the do in each case (e.g. show or hide specific guidance).

    - `on_acceptable(self, task: str)`
    - `on_unacceptable(self, task: str)`
    - `on_active(self, task: str)`
    - `on_inactive(self, task: str)`

    A history of user input events are also recorded and can be conveniently accessed via the `get_latest_user_input(self, event_type: type, n: int = 1) -> Iterator[Event]` method.
    """

    def __init__(
        self,
        sensors: list[Sensor],
        actuators: list[Actuator],
        user_input_events: tuple[type[Event]] = None,
        user_input_events_history_size: int | list[int] = 50,
    ):
        """Constructor.

        Args:
            sensors (list[Sensor]): list of sensors, this will typically be a list of `icua.agent.TaskAcceptabilitySensor`s. A `UserInputSensor` will always be added automatically.
            actuators (list[Actuator]): list of actuators, this will typically contain actuators that are capable of providing visual feedback to a user, see e.g. `icua.agent.GuidanceActuator` and its concrete implementations.
            user_input_events (tuple[type[Event]], optional): additional user input events to subscribe to. Defaults to None.
            user_input_events_history_size (int | list[int], optional): _description_. Defaults to 50.
        """
        # this is the guidance agents main sensor, it will sense:
        # user input events (typically) MouseButtonEvent, MouseMotionEvent, KeyEvent
        user_input_events = user_input_events if user_input_events else []
        user_input_sensor = UserInputSensor(subscribe_to=[*user_input_events])
        user_input_events = user_input_sensor._subscribe_to

        # used when tasks change their acceptability status (receives TaskAcceptable and TaskUnacceptable actions)
        # TODO perhaps we need one for tasks that are active/inactive)
        # but EnableTask and DisableTask may serve this purpose...
        _task_acceptability_actuator = TaskAcceptabilityActuator()
        super().__init__(
            [user_input_sensor, *sensors],
            [_task_acceptability_actuator, *actuators],
        )
        # agent's beliefs store
        self.beliefs = defaultdict(lambda: None)
        # track the acceptability of each task
        self._is_task_acceptable: dict[str, bool] = defaultdict(lambda: False)
        self._is_task_active: dict[str, bool] = defaultdict(lambda: False)

        # set up buffers for storing user input events
        if isinstance(user_input_events_history_size, int):
            user_input_events_history_size = [user_input_events_history_size] * len(
                user_input_events
            )
        self._user_input_events = {
            t: deque(maxlen=hsize)
            for t, hsize in zip(user_input_events, user_input_events_history_size)
        }
        # add an observe method to capture all user input events (based on the UserInputSensor types)
        self.add_observe(self.on_user_input, self.user_input_types)

    def on_acceptable(self, task: str):
        """Called when a task enters an acceptable state.

        Args:
            task (str): the task.
        """
        pass

    def on_unacceptable(self, task: str):
        """Called when a task enters an unacceptable state.

        Args:
            task (str): the task.
        """
        pass

    def on_active(self, task: str):
        """Called when a task is made active, typically this means it has been enabled in the environment state, but otherwise it may be that the task has appeared to the user (e.g. if it was previously hidden or not ready for user interaction).

        Args:
            task (str): the task.
        """
        pass

    def on_inactive(self, task: str):
        """Called when a task is made inactive, typically this means it has been disabled in the environment state, but otherwise it may be that the task is just hidden or not ready for user interaction.

        Args:
            task (str): the task.
        """
        pass

    @property
    def user_input_types(self):
        """The types of user input that this agent is tracking (READ ONLY)."""
        return tuple(self._user_input_events.keys())

    def get_latest_user_input(self, event_type: type, n: int = 1) -> Iterator[Event]:
        """Getter for the lastest user input of a given type.

        Args:
            event_type (type): the type of user input to get.
            n (int, optional): the number of events to retrieve. Defaults to 1 (the latest event).

        Returns:
            Iterator[Event]: an iterator that contains the requested events (may be empty if no such events exist).
        """
        try:
            return islice(self._user_input_events[event_type], 0, n)
        except IndexError:
            return iter([])
        except KeyError:
            return iter([])

    @observe
    def on_error(
        self,
        observation: ErrorActiveObservation | ErrorObservation,
        component: Component,
    ):
        """Called if this agent receives an `star_ray.event.ErrorObservation` from a component. By default this will re-raise the exception that the observation contains. This can be overriden to alter the default behaviour. The overriding method must be decorated with the `@observe` decorator to work correctly. This method should not be called manually and will be handled by this agents event routing mechanism.

        Args:
            observation (ErrorActiveObservation | ErrorObservation): the error observation.
            component (Component): the component that the observation originated from.

        Raises:
            observation.exception: the exception contained in the observation (raised by default).
        """
        raise observation.exception()

    @observe
    def on_task_acceptability(self, observation: TaskAcceptabilityObservation):
        """Called if this agent receives a `icua.agent.TaskAcceptabilityObservation` from one of its sensors.

        This will trigger the relevant callback:
        - `on_acceptable(self, task: str)`
        - `on_unacceptable(self, task: str)`
        - `on_active(self, task: str)`
        - `on_inactive(self, task: str)`

        This method should not be called manually and will be handled by this agents event routing mechanism.

        Args:
            observation (TaskAcceptabilityObservation): the acceptability observation
        """
        task = observation.values["task"]
        was_active = self._is_task_active[task]
        was_acceptable = self._is_task_acceptable[task]
        is_active = observation.values["is_active"]
        is_acceptable = observation.values["is_acceptable"]
        self._is_task_active[task] = is_active
        self._is_task_acceptable[task] = is_acceptable
        if was_active and not is_active:
            self.on_inactive(task)
        elif not was_active and is_active:
            self.on_active(task)
        if was_acceptable and not is_acceptable:
            self.attempt(TaskUnacceptable(task=task))
            self.on_unacceptable(task)
        elif not was_acceptable and is_acceptable:
            self.attempt(TaskAcceptable(task=task))
            self.on_acceptable(task)

    # this is manually added to the event router (see __init__), so no @observe here
    def on_user_input(self, observation: Any):
        """Called when this agent receives a user input event, it will add the event to an internal buffer. See `get_latest_user_input`.

        This method should not be called manually and will be handled by this agents event routing mechanism.

        Args:
            observation (Any): the observation.
        """
        # print("USER INPUT:", observation)
        assert isinstance(observation, self.user_input_types)
        self._user_input_events[type(observation)].appendleft(observation)



import time


__all__ = ("TimeStat",)


class TimeStat:
    def __init__(self):
        self.reset_stats()

    def reset_stats(self):
        self.count = 0
        self.total_time = 0
        self.min_time = float("inf")
        self.max_time = float("-inf")
        self.avg_time = 0

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        end_time = time.time()
        elapsed_time = end_time - self.start_time
        self.count += 1
        self.total_time += elapsed_time

        # Update min and max times
        self.min_time = min(self.min_time, elapsed_time)
        self.max_time = max(self.max_time, elapsed_time)

        # Update the average time
        if self.count > 0:
            self.avg_time = self.total_time / self.count

    def get_stat(self):
        stats = {
            "min_time": self.min_time if self.count > 0 else None,
            "max_time": self.max_time if self.count > 0 else None,
            "avg_time": self.avg_time if self.count > 0 else None,
        }
        # Reset the stats after returning them
        self.reset_stats()
        return stats


if __name__ == "__main__":
    # Example usage
    with TimeStat() as timer:
        # Code block whose execution time you want to measure
        time.sleep(1)

    stats = timer.get_stat()
    print(stats)



"""Module defines task related events: `DisableTask`, `EnableTask`, see class documentation for details."""

from typing import Any
from star_ray import Event
from pydantic import Field

__all__ = ("EnableTask", "DisableTask")


class DisableTask(Event):
    """Event that disables a given task.

    Attributes:
        task_name (str): name of the task to disable.
    """

    task_name: str


class EnableTask(Event):
    """Event that enables a given task.

    Attributes:
        task_name (str): name of the task to enable.
        context (dict[str,Any]): context to use when enabling the task
        insert_at (int): at what position in the SVG tree to insert the task element.
    """

    task_name: str
    context: dict[str, Any] | None = Field(default=None)
    insert_at: int = Field(default=0)



from importlib.resources import files

# TODO remove some of these.. we dont need them!
# directories
DEFAULT_STATIC_PATH = files(__package__).parent.parent / "static"
DEFAULT_TASK_PATH = DEFAULT_STATIC_PATH / "task"
DEFAULT_CONFIG_PATH = DEFAULT_STATIC_PATH / "config"
DEFAULT_SCHEDULE_PATH = DEFAULT_STATIC_PATH / "schedule"
# files
DEFAULT_SVG_INDEX_FILE = DEFAULT_STATIC_PATH / "index.svg.jinja"
DEFAULT_SERVER_CONFIG_FILE = DEFAULT_CONFIG_PATH / "default_server_config.json"
DEFAULT_SCHEDULE_FILE = DEFAULT_SCHEDULE_PATH / "default_schedule.sch"

DEFAULT_SVG_NAMESPACES = {"svg": "http://www.w3.org/2000/svg"}
DEFAULT_SVG = """<svg:svg id="root" x="0" y="0" xmlns:svg="http://www.w3.org/2000/svg"></svg:svg>"""

# TODO this should be set in config file somewhere
DEFAULT_KEY_BINDING = {
    "ArrowRight": "right",
    "ArrowLeft": "left",
    "ArrowUp": "up",
    "ArrowDown": "down",
}

__all__ = (
    "DEFAULT_STATIC_PATH",
    "DEFAULT_TASK_PATH",
    "DEFAULT_CONFIG_PATH",
    "DEFAULT_SCHEDULE_PATH",
    "DEFAULT_SVG_INDEX_FILE",
    "DEFAULT_SERVER_CONFIG_FILE",
    "DEFAULT_SCHEDULE_FILE",
    "DEFAULT_SVG_NAMESPACES",
    "DEFAULT_KEY_BINDING",
)



"""Various utilities."""

from ._error import ICUAInternalError
from ._logging import LOGGER, EventLogger
from ._task_loader import TaskLoader, Task, AvatarFactory
from ._schedule import ScheduledAgent, ScheduledAgentAsync, ScheduledAgentFactory
from ._timestat import TimeStat

__all__ = (
    "TaskLoader",
    "Task",
    "TimeStat",
    "ScheduledAgentFactory",
    "ScheduledAgent",
    "ScheduledAgentAsync",
    "AvatarFactory",
    "ScheduledAgent",
    "ICUAInternalError",
    "LOGGER",
    "EventLogger",
)



"""Module containing the abstract base class `GuidanceActuator` and some concrete implementations: `ArrowGuidanceActuator` and `BoxGuidanceActuator` both provide convenient attempt methods for display visual guidance to a user."""

from abc import abstractmethod
from typing import Literal, Any
from star_ray.agent import Actuator, attempt
from star_ray.event import Action
from icua.event import MouseMotionEvent, EyeMotionEvent
from star_ray.agent.agent import Agent

from ..event.event_guidance import (
    DrawArrowAction,
    DrawBoxOnElementAction,
    ShowElementAction,
    HideElementAction,
    ShowGuidance,
    HideGuidance,
)


class GuidanceActuator(Actuator):
    """Abstract base class for guidance actuators."""

    @abstractmethod
    def show_guidance(self, **kwargs):
        """An attempt method that will show guidance to a user."""

    @abstractmethod
    def hide_guidance(self, **kwargs):
        """An attempt method that will hide guidance from a user."""


class CounterFactualGuidanceActuator(GuidanceActuator):
    """A simple guidance actuator that makes no change to the UI, but will take `ShowGuidance` and `HideGuidance` actions. These actions may be logged an used to determine when an agent decided to show/hide guidance to the user."""

    @attempt()
    def show_guidance(self, task: str) -> list[Action]:  # noqa
        return ShowGuidance(task=task)

    @attempt()
    def hide_guidance(self, task: str) -> list[Action]:  # noqa
        return HideGuidance(task=task)


class BoxGuidanceActuator(GuidanceActuator):
    """A concrete implementation of `GuidanceActuator` that implements box guidance. The box bounds a given task element serving to highlight it to the user, typically the task will be one that is not in an acceptible state."""

    def __init__(
        self,
        box_stroke_color: str = "#ff0000",
        box_stroke_width: float = 4.0,
    ):
        """Constructor.

        Args:
            box_stroke_color (str, optional): color of the box. Defaults to "#ff0000".
            box_stroke_width (float, optional): width of the box outline. Defaults to 4.0.
        """
        super().__init__()
        self._box_stroke_color = box_stroke_color
        self._box_stroke_width = box_stroke_width
        self._guidance_box_id_template = "guidance_box_%s"
        # ids of each of the guidance boxes that have been created TODO this should be removed when `on_remove` is called!
        self._guidance_boxes = set()

    def on_remove(self, agent: Agent) -> None:  # noqa
        return super().on_remove(agent)  # TODO remove all guidance boxes!

    @attempt()
    def draw_guidance_box_on_element(
        self, element_id: str, **box_data: dict[str, Any]
    ) -> DrawBoxOnElementAction:
        """Attempt method that will draw a box around a given element.

        Args:
            element_id (str): the `id` of the element.
            box_data (dict[str,Any]): data associated with the box.

        Returns:
            DrawBoxOnElementAction: action
        """
        # TODO explicit parameters for box data
        box_data["id"] = self._guidance_box_id_template % element_id
        return DrawBoxOnElementAction(
            xpath=f"//*[@id='{element_id}']", box_data=box_data
        )

    @attempt()
    def show_guidance(self, task: str) -> list[Action]:
        """Show guidance on the given task.

        Args:
            task (str): task to show guidance for.

        Returns:
            list[Action]: guidance actions
        """
        self._guidance_on = task
        guidance_box_id = self._guidance_box_id_template % task

        actions = []

        if guidance_box_id not in self._guidance_boxes:
            # first time! insert the guidance box
            self._guidance_boxes.add(guidance_box_id)
            box_data = {
                "stroke-width": self._box_stroke_width,
                "stroke": self._box_stroke_color,
            }
            # draw the box but it is hidden (opacity=0)
            actions.append(
                self.draw_guidance_box_on_element(task, opacity=0.0, **box_data)
            )
        actions.append(ShowElementAction(xpath=f"//*[@id='{guidance_box_id}']"))
        return actions

    @attempt()
    def hide_guidance(self, task: str):
        """Hide guidance on the given task.

        Args:
            task (str): task to hide guidance for.

        Returns:
            list[Action]: guidance actions
        """
        self._guidance_on = None
        guidance_box_id = self._guidance_box_id_template % task

        actions = []

        if guidance_box_id not in self._guidance_boxes:
            # first time! insert the guidance box
            self._guidance_boxes.add(guidance_box_id)
            box_data = {
                "stroke-width": self._box_stroke_width,
                "stroke": self._box_stroke_color,
            }
            # draw the box but it is hidden (opacity=0)
            actions.append(
                self.draw_guidance_box_on_element(task, opacity=0.0, **box_data)
            )
        actions.append(HideElementAction(xpath=f"//*[@id='{guidance_box_id}']"))
        return actions


class ArrowGuidanceActuator(GuidanceActuator):
    """A concrete implementation of `GuidanceActuator` that implements a guidance arrow. The arrow is displayed at the users mouse (or gaze) position that points towards a given task element, typically this task will be one that is not in an acceptible state."""

    ARROW_MODES = Literal["gaze", "mouse", "fixed"]

    def __init__(
        self,
        arrow_mode: Literal["gaze", "mouse", "fixed"],
        arrow_scale: float = 1.0,
        arrow_fill_color: str = "none",
        arrow_stroke_color: str = "#ff0000",
        arrow_stroke_width: float = 4.0,
        arrow_offset: tuple[float, float] = (80, 80),
    ):
        """Constructor.

        Args:
            arrow_mode (Literal): modes for arrow display,
            arrow_scale (float, optional): scale of the arrow. Defaults to 1.0.
            arrow_fill_color (str, optional): fill colour of the arrow. Defaults to "none".
            arrow_stroke_color (str, optional): line colour of the arrow outlien. Defaults to "#ff0000".
            arrow_stroke_width (float, optional): line width of the arrow outline. Defaults to 4.0.
            arrow_offset (tuple[float, float], optional): offset of the arrow from its set position. Defaults to (80, 80).

        Raises:
            ValueError: _description_
        """
        super().__init__()
        self._arrow_mode = arrow_mode
        self._arrow_scale = arrow_scale
        self._arrow_fill_color = arrow_fill_color
        self._arrow_stroke_color = arrow_stroke_color
        self._arrow_stroke_width = arrow_stroke_width
        self._arrow_offset = arrow_offset

        if self._arrow_mode not in ArrowGuidanceActuator.ARROW_MODES.__args__:
            raise ValueError(
                f"Invalid argument: `arrow_mode` must be one of {ArrowGuidanceActuator.ARROW_MODES}"
            )
        self._guidance_arrow_id = "guidance_arrow"
        self._guidance_on = None
        self._gaze_position = None
        self._mouse_position = None

    def __attempt__(self):  # noqa
        if self._guidance_on is None:
            return []
        elif self._arrow_mode == "gaze":
            if self._gaze_position:  # TODO check that this doesnt continuously hold...?
                attrs = dict(
                    id=self._guidance_arrow_id,
                    x=self._gaze_position[0] + self._arrow_offset[0],
                    y=self._gaze_position[1] + self._arrow_offset[0],
                    point_to=self._guidance_on,
                )
                return [DrawArrowAction(xpath="/svg:svg", data=attrs)]
            return []
        elif self._arrow_mode == "mouse":
            if self._mouse_position:
                attrs = dict(
                    id=self._guidance_arrow_id,
                    x=self._mouse_position[0] + self._arrow_offset[0],
                    y=self._mouse_position[1] + self._arrow_offset[0],
                    point_to=self._guidance_on,
                )
                return [DrawArrowAction(xpath="/svg:svg", data=attrs)]
            return []
        elif self._arrow_mode == "fixed":
            # TODO where should it be? the center of the screen?
            raise NotImplementedError("TODO")
        else:
            raise ValueError(
                f"Invalid argument: `arrow_mode` must be one of {ArrowGuidanceActuator.ARROW_MODES}"
            )

    @attempt([EyeMotionEvent])
    def set_gaze_position(self, action: EyeMotionEvent) -> None:
        """Sets the users current gaze position. This may be used as a position for arrow display."""
        self._gaze_position = action.position

    @attempt([MouseMotionEvent])
    def set_mouse_motion(self, action: MouseMotionEvent) -> None:
        """Sets the users current mouse position. This may be used as a position for arrow display."""
        self._mouse_position = action.position

    def on_add(self, agent: Agent) -> None:  # noqa
        super().on_add(agent)
        self.draw_guidance_arrow(
            self._guidance_arrow_id,
            0.0,
            0.0,
            fill=self._arrow_fill_color,
            stroke_width=self._arrow_stroke_width,
            stroke_color=self._arrow_stroke_color,
            opacity=0.0,
            scale=self._arrow_scale,
        )

    def on_remove(self, agent: Agent) -> None:  # noqa
        super().on_remove(agent)
        # TODO remove arrow element!

    @attempt()
    def draw_guidance_arrow(
        self,
        name: str,
        x: float,
        y: float,
        scale: float = 1.0,
        rotation: float = 0.0,
        fill: str = "none",
        opacity: float = 0.0,
        stroke_color: str = "#ff0000",
        stroke_width: float = 2.0,
        **kwargs,
    ) -> DrawArrowAction:
        """Attempt method that takes an action to draw a guidance arrow.

        Args:
            name (str): id.
            x (float): x position.
            y (float): y position.
            scale (float, optional): scale. Defaults to 1.0.
            rotation (float, optional): rotation. Defaults to 0.0.
            fill (str, optional): fill color. Defaults to "none".
            opacity (float, optional): opacity (0.0 means hidden). Defaults to 0.0.
            stroke_color (str, optional): color of the arrow border. Defaults to "#ff0000".
            stroke_width (float, optional): thickness of the arrow border. Defaults to 2.0.
            kwargs : (dict[str,Any]): additional optional keyword arguments.

        Returns:
            DrawArrowAction: action
        """
        return DrawArrowAction(
            xpath="/svg:svg",
            data=dict(
                id=name,
                x=x,
                y=y,
                scale=scale,
                rotation=rotation,
                fill=fill,
                opacity=opacity,
                stroke=stroke_color,
                stroke_width=stroke_width,
                **kwargs,
            ),
        )

    @attempt()
    def show_guidance(self, task: str):
        """Show guidance on the given task.

        Args:
            task (str): task to show guidance for.

        Returns:
            list[Action]: guidance actions
        """
        self._guidance_on = task
        actions = [
            ShowElementAction(xpath=f"//*[@id='{self._guidance_arrow_id}']"),
        ]
        return actions

    @attempt()
    def hide_guidance(self, task: str):
        """Hide guidance on the given task.

        Args:
            task (str): task to hide guidance for.

        Returns:
            list[Action]: guidance actions
        """
        self._guidance_on = None
        actions = [
            HideElementAction(xpath=f"//*[@id='{self._guidance_arrow_id}']"),
        ]
        return actions



"""Get various kinds of data from parsed event logs."""

from collections.abc import Callable
import pandas as pd
import inspect
from star_ray import Event
from icua.event import (
    ShowGuidance,
    HideGuidance,
    TaskAcceptable,
    TaskUnacceptable,
)


def dedup(df: pd.DataFrame, col: str) -> pd.DataFrame:
    """Remove consective duplicates from df by the value in `col`.

    Args:
        df (pd.DataFrame): dataframe to remove values from.
        col (str): column to use

    Returns:
        pd.DataFrame: the dedupped dataframe
    """
    return df[df[col] != df[col].shift()]


def _generate_task_intervals(
    events: list[tuple[float, Event]],
    start: type | Callable[[Event], bool],
    end: type | Callable[[Event], bool],
):
    start_time, end_time = events[0][0], events[-1][0]

    if inspect.isclass(start):

        def start_check(x, _start=start):
            return isinstance(x, _start)

        start_check.__name__ = start.__name__
        start = start_check
    if inspect.isclass(end):

        def end_check(x, _end=end):
            return isinstance(x, _end)

        end_check.__name__ = end.__name__
        end = end_check

    def _gen():
        for t, event in events:
            if start(event):
                yield t, event.task, start.__name__
            elif end(event):
                yield t, event.task, end.__name__

    for task, df in pd.DataFrame(_gen(), columns=["timestamp", "task", "type"]).groupby(
        "task"
    ):
        # print(task, df)
        # remove consective duplicates, these are meaningless (there shouldnt be any...?)
        df = dedup(df.reset_index(drop=True), col="type").reset_index(drop=True)
        # insert events if needed
        if df["type"].iloc[0] != start.__name__:
            df = pd.concat(
                [pd.DataFrame([{"timestamp": start_time, "type": start.__name__}]), df]
            ).reset_index(drop=True)

        if df["type"].iloc[-1] != end.__name__:
            df = pd.concat(
                [df, pd.DataFrame([{"timestamp": end_time, "type": end.__name__}])]
            ).reset_index(drop=True)

        assert len(df) % 2 == 0  # must be an event number to create intervals!
        intervals = df["timestamp"].to_numpy().reshape(-1, 2)
        yield task, intervals


# def get_mouse_click_intervals(events: list[tuple[float, Event]]):
#     def mouse_down(event: MouseButtonEvent):
#         return event.status == MouseButtonEvent.DOWN

#     def mouse_up(event: MouseButtonEvent):
#         return event.status == MouseButtonEvent.UP

#     yield from _generate_task_intervals(events, mouse_down, mouse_up)


# def get_mouse_down_on(events: list[tuple[float, Event]], element: str) -> np.ndarray:
#     events = filter(
#         lambda x: isinstance(x[1], MouseButtonEvent) and element in x[1].target, events
#     )
#     return np.array([e[0] for e in events])


def get_guidance_intervals(events: list[tuple[float, Event]]):
    """Generator that will get guidance intervals by task. The intervals always start with "ShowGuidance" and end with "HideGuidance".

    Args:
        events (list[tuple[float, Event]]): events from event log

    Yields:
        (str, np.ndarray): task name, intervals
    """
    start, end = ShowGuidance, HideGuidance
    yield from _generate_task_intervals(events, start, end)


def get_acceptable_intervals(events: list[tuple[float, Event]]):
    """Generator that will get acceptability intervals by task. The intervals always start with "TaskAcceptable" and end with "TaskUnacceptable".

    Args:
        events (list[tuple[float, Event]]): events from event log

    Yields:
        (str, np.ndarray): task name, intervals
    """
    start, end = TaskAcceptable, TaskUnacceptable
    yield from _generate_task_intervals(events, start, end)


if __name__ == "__main__":
    # TODO move this to a test file...
    from pydantic import BaseModel

    class A(BaseModel):  # noqa
        task: str

    class B(BaseModel):  # noqa
        task: str

    def _events():
        for i in range(10):
            yield i, A(task="task") if i % 2 == 0 else B(task="task")

    events = list(_events())
    for task, intervals in _generate_task_intervals(events, A, B):
        print(task, intervals)

    def _events():
        for i in range(10):
            yield i, A(task="task") if (i + 1) % 2 == 0 else B(task="task")

    events = list(_events())
    for task, intervals in _generate_task_intervals(events, A, B):
        print(task, intervals)



"""Module defines the `EyetrackerIOSensor` class which may be attached to an agent to receive `EyeMotionEvent`s as user input, see class documentation for details."""

from star_ray.agent import IOSensor, attempt
from .eyetrackerbase import EyetrackerBase
from .filter import IVTFilter, WindowSpaceFilter, NWMAFilter
from .event import EyeMotionEventRaw, EyeMotionEvent
from ...event import WindowResizeEvent, WindowMoveEvent, ScreenSizeEvent


class EyetrackerIOSensor(IOSensor):
    """An `IOSensor` implementation that gathers observations `EyeMotionEvent` from an eyetracker."""

    def __init__(
        self,
        eyetracker: EyetrackerBase,
        velocity_threshold: float = 0.1,
        moving_average: int = 10,
    ):
        """Constructor.

        Args:
            eyetracker (EyetrackerBase): base eyetracker, this is the IO device of this `IOSensor`.
            velocity_threshold (float, optional): velocity threshold for the `IVTFilter`. Defaults to 0.1.
            moving_average (int, optional): moving average window size for the `NWMAFilter`. Defaults to 10.
        """
        super().__init__(eyetracker)
        self._ivt_filter = IVTFilter(velocity_threshold)
        nan = tuple([float("nan"), float("nan")])
        self._ws_filter = WindowSpaceFilter(nan, nan, nan)
        self._ma_filter = NWMAFilter(moving_average)

    def __transduce__(self, events: list[EyeMotionEventRaw]) -> list[EyeMotionEvent]:
        """Converts the list of raw eyetracking events to a list of eyetracking events by appling the filters that are part of this sensor.

        The filters are applied in order:
        - `NWMAFilter` - computes a moving average over points
        - `IVTFilter` - computes fixation/saccade based on velocity threshold.
        - `WindowSpaceFilter` - computes window UI space coordinates from screen space (requires screen & window information, see methods: `EyetrackerIOSensor.on_window_move`, `EyetrackerIOSensor.on_window_resize`, `EyetrackerIOSensor.on_screen_size`).

        Args:
            events (list[EyeMotionEventRaw]): raw eyetracking events

        Returns:
            list[EyeMotionEvent]: eyetracking events
        """
        return list(self._transduce_iter(events))

    def _transduce_iter(self, events: list[EyeMotionEventRaw]):  # noqa
        # applies all filters to the each eye motion event
        for event in events:
            data = event.model_dump()
            for filter in [self._ma_filter, self._ivt_filter, self._ws_filter]:
                data = filter(data)
            # NOTE: position needs to be set properly in the agent (i.e. convert to view space)
            data["position_raw"] = data["position"]
            yield EyeMotionEvent.model_validate(data)

    @attempt
    def on_window_move(self, event: WindowMoveEvent) -> None:
        """Should be called when the UI window is moved. It is the responsibility of the attached agent to provide this information to this sensor. It is required to transform raw eyetracking coordinates to UI coordinates.

        If this sensor is attached to a `AgentRouted` this may be called automatically.

        Args:
            event (WindowMoveEvent): event with information about the UI window position.
        """
        # print("WINDOW MOVE", event.position)
        self._ws_filter.window_position = event.position

    @attempt
    def on_window_resize(self, event: WindowResizeEvent) -> None:
        """Should be called when the UI window is resized. It is the responsibility of the attached agent to provide this information to this sensor. It is required to transform raw eyetracking coordinates to UI coordinates.

        If this sensor is attached to a `AgentRouted` this may be called automatically.

        Args:
            event (WindowResizeEvent): event with information about the UI window size.
        """
        # print("WINDOW RESIZE!", event.size)
        self._ws_filter.window_size = event.size

    @attempt
    def on_screen_size(self, event: ScreenSizeEvent) -> None:
        """Should be called when the screen/monitor size is discovered for the first time. It is the responsibility of the attached agent to provide this information to this sensor. It is required to transform raw eyetracking coordinates to UI coordinates.

        If this sensor is attached to a `AgentRouted` this may be called automatically.

        Args:
            event (ScreenSizeEvent): event with information about the screen/monitor size.
        """
        # print("SCREEN SIZE!", event.size)
        self._ws_filter.screen_size = event.size



"""TODO."""
# from pydantic import BaseModel
# from abc import ABC, abstractmethod
# from star_ray import Agent, Actuator
# from typing import List, Type, Callable

# from ..utils._schedule import ScheduledAgent

# class _Task(BaseModel, ABC):
#     name: str

#     @abstractmethod
#     def get_agent(self, agent: Agent | None = None):
#         pass

#     @abstractmethod
#     def get_avatar(self, avatar: Agent | None = None):
#         pass


# class _ScheduledTask(_Task):

#     agent_actuators: List[Type[Actuator]]
#     avatar_actuators: List[Type[Actuator]]

#     def get_agent(self, agent: Agent | Type[Agent] = None):
#         if agent is None:

#         return _ScheduledTask._initialise_agent(agent, self.agent_actuators)

#     def get_avatar(self, avatar: Agent | Type[Agent] | Callable[[], Agent]):
#         return _ScheduledTask._initialise_agent(avatar, self.avatar_actuators)

#     @staticmethod
#     def _initialise_agent(
#         agent: Agent | Type[Agent] | Callable[[], Agent],
#         actuators: List[Type[Actuator]],
#     ):
#         if isinstance(agent, Agent):
#             for actuator in actuators:
#                 agent.add_component(actuator())
#             return agent
#         else:  # presume this is a type (or factory method)
#             return agent(sensors=[], actuators=[actuator() for actuator in actuators])


# class _TaskDynamicLoad(_Task):
#     path: str



"""Package containing tools for post-experiment analysis."""

from .event_log_parser import EventLogParser
from .plot import (
    plot_timestamps,
    plot_intervals,
)
from .get import get_guidance_intervals, get_acceptable_intervals

__all__ = (
    "EventLogParser",
    "plot_timestamps",
    "plot_intervals",
    "get_guidance_intervals",
    "get_acceptable_intervals",
)



"""Module defining the `MultiTaskEnvironment`, see class documentation for details."""

import asyncio
from typing import Any
from collections.abc import Callable

from star_ray import Environment, Agent, Actuator
from .multitask_ambient import MultiTaskAmbient

from ..utils import LOGGER


class MultiTaskEnvironment(Environment):
    """Environment implementation that supports "tasks". A task is a modular part of the environment which the user typically interacts with via their `Avatar`. Tasks will typially have an associated agent that updates/manages the tasks state. The users avatar should also provide some means to interact with tasks. When a new task is added, this may mean the avatar gains a new actuator and will often mean that a new agent is added to the environment.

    TODO a thorough explaination of what tasks are and how to define/use them!

    See:
        `MultiTaskAmbient`
    """

    def __init__(
        self,
        avatar: Agent,
        agents: list[Agent] = None,
        wait: float = 0.01,
        svg_size: tuple[float, float] = None,
        svg_position: tuple[float, float] = None,
        logging_path: str = None,
        terminate_after: float = -1,
        **kwargs: dict[str, Any],
    ):
        """Constructor.

        Args:
            avatar (Agent, optional): The users avatar. Defaults to None.
            agents (list[Agent], optional): list of initial agents. Defaults to None.
            wait (float, optional): time to wait between simulation cycles. Defaults to 0.01.
            svg_size (tuple[float, float], optional): size of the root SVG element. Defaults to None (see `MultiTaskAmbient` for details).
            svg_position (tuple[float, float], optional): position of the root SVG element. Defaults to None (see `MultiTaskAmbient` for details).
            logging_path (str, optional): path that events will be logged to. Defaults to None (see `MultiTaskAmbient` for details).
            terminate_after (float, optional): time after which to terminate the simulatio. Defaults to -1 (never terminate).
            kwargs (dict[str, Any]): Additional optional keyword arguments, see `MultiTaskAmbient` for options.
        """
        ambient = MultiTaskAmbient(
            avatar=avatar,
            agents=agents,
            logging_path=logging_path,
            svg_size=svg_size,
            svg_position=svg_position,
            **kwargs,
        )
        super().__init__(
            ambient=ambient,
            wait=wait,
            sync=True,
        )
        # time until termination
        self._terminate_after = terminate_after

    @property
    def ambient(self) -> MultiTaskAmbient:
        """Getter for the inner ambient, which is always a `MultiTaskAmbient`. IMPORTANT NOTE: remote ambients are not currently supported.

        Returns:
            MultiTaskAmbient: the ambient.
        """
        ambient = self._ambient._inner
        # TODO remote ambient currently not supported - this would be easy to do... (just need relevant methods exposed)
        assert isinstance(ambient, MultiTaskAmbient)
        return ambient

    def enable_task(
        self, task_name: str, context: dict[str, Any] = None, insert_at: int = -1
    ):
        """Manually enable a task. Tasks may otherwise be enabled via an `EnableTask` action.

        Args:
            task_name (str): name of the task to enable.
            context (dict[str,Any]): context to use when enabling the task
            insert_at (int): at what position in the SVG tree to insert the task element.
        """
        self.ambient.enable_task(task_name, context=context, insert_at=insert_at)

    def disable_task(self, task_name: str):
        """Manually disable a task. Tasks may otherwise be disabled via an `EnableTask` action.

        Args:
            task_name (str): name of the task to disable.
        """
        self.ambient.disable_task(task_name)

    def rename_task(self, task_name: str, new_name: str) -> None:
        """Rename a task. If the task is enabled this will alter the `id` attribute of the task element (TODO).

        Args:
            task_name (str): current name of the task.
            new_name (str): new name of the task.
        """
        return self.ambient.rename_task(task_name, new_name)

    def add_task(
        self,
        name: str,
        path: str,
        agent_actuators: list[Callable[[], Actuator]] = None,
        avatar_actuators: list[Callable[[], Actuator]] = None,
        enable: bool = False,
    ):
        """Add a new task, this will load all required files and prepare the task but will not start the task (unless `enable` is True). To start the task use `enable_task` or have an agent take the `EnableTask` action.

        Args:
            name (str): the unique name of the task.
            path (str | list[str]): path(s) to task files.
            agent_actuators (list[Callable[[], Actuator]] | None, optional): actuators with which to create agents from a schedule file, see `MultiTaskEnvironment` documentation for details. Defaults to None.
            avatar_actuators (list[Callable[[], Actuator]] | None, optional): actuators that will be added to the avatar upon enabling the task, see `MultiTaskEnvironment` documentation for details. Defaults to None.
            enable (bool, optional): whether to immediately enable the task. Defaults to False.
        """
        # attempt to log the files that will be used in the task
        self.ambient.add_task(
            name,
            path,
            agent_actuators=agent_actuators,
            avatar_actuators=avatar_actuators,
            enable=enable,
        )

    async def __initialise__(self, event_loop):  # noqa
        await self._ambient.__initialise__()

    def get_schedule(self) -> list[asyncio.Task]:  # noqa
        tasks = super().get_schedule()
        # whether to terminate after a given period of time
        if self._terminate_after > 0:

            async def _terminate_after(env: MultiTaskEnvironment):
                try:
                    await asyncio.sleep(env._terminate_after)
                    # TODO perhaps this should be an event that is logged?
                    LOGGER.debug(
                        f"Closing simulation: time limit ({env._terminate_after}s) reached"
                    )
                    await env.ambient.__terminate__()
                except asyncio.CancelledError:
                    pass
                except asyncio.TimeoutError:
                    pass

            tasks.append(asyncio.create_task(_terminate_after(self)))
        return tasks

    # def run(self):  # noqa
    #     async def _cancel_pending_tasks(pending_tasks, timeout=0.5):
    #         """Await pending tasks with a timeout to avoid hanging."""
    #         for pending_task in pending_tasks:
    #             pending_task.cancel()
    #         for pending_task in pending_tasks:
    #             try:
    #                 await asyncio.wait_for(pending_task, timeout=timeout)
    #             except asyncio.CancelledError:
    #                 pass  # print(f"Pending task {pending_task} was cancelled")
    #             except asyncio.TimeoutError:
    #                 LOGGER.warning(
    #                     f"Pending task {pending_task} did not finish within timeout"
    #                 )
    #                 pass

    #     async def _run():
    #         """Run the simulation. This is a custom implementation of run which may be changed in favour of proper scheduling in future versions (if `star_ray` implements support for this)."""
    #         event_loop = asyncio.get_event_loop()
    #         await self.__initialise__(event_loop)
    #         tasks = self.get_schedule()
    #         # The default behaviour here is to exit the program if there is an exception,
    #         # this is because we dont want any silent failures during an experiment as it may invalidate any experimental results.
    #         # if this is not desired behaviour, you should override this method and continue execution after e.g. logging the exception.
    #         done, pending = await asyncio.wait(
    #             tasks, return_when=asyncio.FIRST_EXCEPTION
    #         )
    #         await _cancel_pending_tasks(pending)
    #         for task in done:
    #             e = task.exception()
    #             if e:
    #                 raise e

    #     asyncio.run(_run())

    # def get_schedule(self):  # noqa
    #     tasks = []
    #     for agent in self._ambient.get_agents():
    #         if isinstance(agent.get_inner(), ScheduledAgent):
    #             # schedule agents are fully async, they will wait to execute according to their schedule.
    #             tasks.append(asyncio.create_task(self._run_agent_no_wait(agent)))
    #         else:
    #             tasks.append(asyncio.create_task(self._run_agent(agent)))
    #     return tasks

    # async def _run_agent_no_wait(self, agent) -> None:
    #     """Runs an agent without waiting. This should only be used for agents that have async cycle methods to prevent event loop hogging."""
    #     # TODO check that the agent is alive...
    #     try:
    #         while self._ambient.is_alive:  # check that the agent is alive...?
    #             await agent.__sense__(self._ambient)
    #             await agent.__cycle__()
    #             await agent.__execute__(self._ambient)
    #     except asyncio.CancelledError:
    #         pass
    #     except Exception as e:
    #         raise e  # re-raise

    # async def _run_agent(self, agent) -> None:
    #     """Run an agent in the usual way waiting for a short time before executing each cycle."""
    #     try:
    #         # TODO check that the agent is alive...
    #         while self._ambient.is_alive:  # check that the agent is alive...?
    #             await asyncio.sleep(self._wait / self._ambient.get_agent_count())
    #             await agent.__sense__(self._ambient)
    #             await agent.__cycle__()
    #             await agent.__execute__(self._ambient)
    #     except asyncio.CancelledError:
    #         pass
    #     except Exception as e:
    #         raise e  # re-raise



"""Bridge between the Tobii (pro) eyetracking API and the `icua` eyetracking API."""

import time
import traceback
from queue import Queue
from .error import EyetrackingConfigurationError
from .eyetrackerbase import EyetrackerBase
from ...utils import LOGGER
from .event import EyeMotionEventRaw
from star_ray import Event

try:
    import tobii_research as _tr  # noqa

    TOBII_RESEACH_SDK_AVALIABLE = True
except ModuleNotFoundError:
    LOGGER.warning("Failed to locate module: `tobii_research`")
    TOBII_RESEACH_SDK_AVALIABLE = False


# constants used by tobii api for raw gaze events
DEVICE_TIME_STAMP = "device_time_stamp"
SYSTEM_TIME_STAMP = "system_time_stamp"
LEFT_GAZE_POINT_ON_DISPLAY_AREA = "left_gaze_point_on_display_area"
LEFT_GAZE_POINT_IN_USER_COORDINATE_SYSTEM = "left_gaze_point_in_user_coordinate_system"
LEFT_GAZE_POINT_VALIDITY = "left_gaze_point_validity"
LEFT_PUPIL_DIAMETER = "left_pupil_diameter"
LEFT_PUPIL_VALIDITY = "left_pupil_validity"
LEFT_GAZE_ORIGIN_IN_USER_COORDINATE_SYSTEM = (
    "left_gaze_origin_in_user_coordinate_system"
)
LEFT_GAZE_ORIGIN_IN_TRACKBOX_COORDINATE_SYSTEM = (
    "left_gaze_origin_in_trackbox_coordinate_system"
)
LEFT_GAZE_ORIGIN_VALIDITY = "left_gaze_origin_validity"
RIGHT_GAZE_POINT_ON_DISPLAY_AREA = "right_gaze_point_on_display_area"
RIGHT_GAZE_POINT_IN_USER_COORDINATE_SYSTEM = (
    "right_gaze_point_in_user_coordinate_system"
)
RIGHT_GAZE_POINT_VALIDITY = "right_gaze_point_validity"
RIGHT_PUPIL_DIAMETER = "right_pupil_diameter"
RIGHT_PUPIL_VALIDITY = "right_pupil_validity"
RIGHT_GAZE_ORIGIN_IN_USER_COORDINATE_SYSTEM = (
    "right_gaze_origin_in_user_coordinate_system"
)
RIGHT_GAZE_ORIGIN_IN_TRACKBOX_COORDINATE_SYSTEM = (
    "right_gaze_origin_in_trackbox_coordinate_system"
)
RIGHT_GAZE_ORIGIN_VALIDITY = "right_gaze_origin_validity"


class TobiiEyetracker(EyetrackerBase):
    """This class exposes the Tobii (pro) eyetracking API. It may be used by an avatar (or the environment) to get data on a users gaze location."""

    def __init__(
        self,
        uri: str | None = None,
        **kwargs,
    ):
        """Constructor.

        Args:
            uri (str): the unique address of the eyetracker hardware, this can be found in the `Tobii Pro Eyetracker Manager` example: "tet-tcp://172.28.195.1". This software should also be used to calibrate the eyetracker, there is currently no built in calibrating from within the icua system. Defaults to None, in which case the first avaliable eyetracker will be used.
            kwargs (dict[str,Any]): additional optional keyword arguments.

        Raises:
            ModuleNotFoundError: If the required `tobii_research` module could not be found.
        """
        # the module was not found, but someone is trying to create an instance of this class!
        try:
            import tobii_research as _  # noqa
        except ModuleNotFoundError as e:
            raise e

        super().__init__()

        self._buffer = Queue()
        self._uri = uri
        self._eyetracker = None
        self._t0 = None

        if uri is None:
            eyetrackers = _tr.find_all_eyetrackers()
            if eyetrackers:
                self._eyetracker = eyetrackers[0]
            else:
                raise EyetrackingConfigurationError(
                    "No eyetrackers were found, check that the eyetracker is connected and avaliable in the official Tobii Pro Eye Tracker Manager software."
                )
        else:
            try:
                self._eyetracker = _tr.EyeTracker(self._uri)
            except Exception as exception:
                raise EyetrackingConfigurationError(
                    f"Failed to initialise eyetracker at uri: {self._uri}, see cause above. Also check that the eyetracker is connected and avaliable in the official Tobii Pro Eye Tracker Manager software. "
                ) from exception

    def start(self):  # noqa
        try:
            # so we can track the actual times (this will introduce a small error...)
            self._t0 = (_tr.get_system_time_stamp(), time.time())
            self._eyetracker.subscribe_to(
                _tr.EYETRACKER_GAZE_DATA,
                self._internal_callback,
                as_dictionary=True,
            )
        except Exception as exception:
            LOGGER.exception(f"Tobii Eyetracker {self._uri} failed to start.")
            # TODO do we want to stop execution when this happens? maybe give an option to continue?
            raise exception
        LOGGER.info(f"Tobii Eyetracker {self._uri} created successfully.")

    def stop(self):  # noqa
        self._eyetracker.unsubscribe_from(
            _tr.EYETRACKER_GAZE_DATA, self._internal_callback
        )

    def _internal_callback(self, gaze_sample) -> None:
        try:
            dt = (gaze_sample[SYSTEM_TIME_STAMP] - self._t0[0]) / 10e5
            timestamp = self._t0[1] + dt
            point = None
            if (
                gaze_sample[LEFT_GAZE_ORIGIN_VALIDITY]
                and gaze_sample[RIGHT_GAZE_POINT_VALIDITY]
            ):
                left_sample = gaze_sample[LEFT_GAZE_POINT_ON_DISPLAY_AREA]
                right_sample = gaze_sample[RIGHT_GAZE_POINT_ON_DISPLAY_AREA]
                point = (
                    (left_sample[0] + right_sample[0]) / 2,
                    (left_sample[1] + right_sample[1]) / 2,
                )
            elif gaze_sample[LEFT_GAZE_POINT_VALIDITY]:
                point = gaze_sample[LEFT_GAZE_POINT_ON_DISPLAY_AREA]
            elif gaze_sample[RIGHT_GAZE_POINT_VALIDITY]:
                point = gaze_sample[RIGHT_GAZE_POINT_ON_DISPLAY_AREA]
            else:
                # it is useful to know when eyetracking events failed
                point = (float("nan"), float("nan"))
            event = EyeMotionEventRaw(timestamp=timestamp, position=point)
            # this is called from another thread (managed by tobii_research, it needs to happen in a thread safe way)
            self._buffer.put_nowait(event)
            # asyncio.get_event_loop().call_soon_threadsafe(
            #    self._buffer.put_nowait, event
            # )
        except Exception as e:
            traceback.print_exception(e)

    async def get(self) -> list[Event]:  # noqa
        items = []
        items.append(await self._buffer.get())
        while not self._buffer.empty():
            item = self._buffer.get_nowait()
            items.append(item)
        return items

    def get_nowait(self) -> list[Event]:  # noqa
        items = []
        while not self._buffer.empty():
            item = self._buffer.get_nowait()
            items.append(item)
        return items



"""Module contains the `MultiTaskAmbient` class, see class documentation for details."""

from typing import Any
from collections.abc import Callable
from star_ray.event import wrap_observation
from star_ray.agent import Agent, Actuator
from star_ray.event import Event, ActiveObservation, ErrorActiveObservation
from star_ray_xml import Insert, Select, XPathElementsNotFound
from star_ray_pygame import SVGAmbient

from ..event import (
    EyeMotionEvent,
    EyeMotionEventRaw,
    MouseButtonEvent,
    MouseMotionEvent,
    KeyEvent,
    WindowCloseEvent,
    WindowOpenEvent,
    WindowFocusEvent,
    WindowMoveEvent,
    WindowResizeEvent,
    ScreenSizeEvent,
)
from ..event import (
    EnableTask,
    DisableTask,
    RenderEvent,
    ShowGuidance,
    HideGuidance,
    TaskUnacceptable,
    TaskAcceptable,
)
from ..utils import TaskLoader, Task
from ..utils._logging import EventLogger


# these actions dont have an effect, but they are important for logging
INERT_ACTIONS = (
    RenderEvent,
    ShowGuidance,
    HideGuidance,
    TaskUnacceptable,
    TaskAcceptable,
)


class MultiTaskAmbient(SVGAmbient):
    """This class manages an SVG state and a collection of tasks that a user may want to interact with. It requires that a special agent the `Avatar` is provided which will act as an interface between the user and the environment. For details on what a `Task` is and how they are used see the `MultiTaskEnvironment` class documentation or read the `icua` wiki page on tasks (TODO provide a link to this)."""

    def __init__(
        self,
        avatar: Agent = None,
        agents: list[Agent] = None,
        svg_size: tuple[float, float] = None,
        svg_position: tuple[float, float] = None,
        logging_path: str = None,
        **kwargs,
    ):
        """Constructor.

        Args:
            avatar (Agent, optional): The users avatar. Defaults to None.
            agents (list[Agent], optional): list of initial agents. Defaults to None.
            svg_size (tuple[float, float], optional): size of the root SVG element, typically this should encompase the bounds of all tasks. Defaults to None, which will use the default `SVGAmbient.DEFAULT_SVG_SIZE`.
            svg_position (tuple[float, float], optional): position of the root SVG element in the UI window. Defaults to None, which which will use the default `SVGAmbient.DEFAULT_SVG_POSITION`.
            logging_path (str, optional): path that events will be logged to. Defaults to None  (see `MultiTaskAmbient` for details).
            kwargs (dict[str, Any]): Additional optional keyword arguments, see `SVGAmbient` for options.
        """
        # this needs to happen before any call to __update__
        self._event_logger = None
        self._initialise_logging(logging_path=logging_path)
        # initialise agents
        agents = agents if agents else []
        agents.append(avatar)
        super().__init__(
            agents,
            svg_size=svg_size,
            svg_position=svg_position,
            **kwargs,
        )
        self._avatar = avatar
        self._task_loader = TaskLoader()
        self._tasks: dict[str, Task] = dict()

    def _initialise_logging(self, logging_path: str = None):
        if logging_path:
            logging_file = EventLogger.default_log_path(
                path=logging_path, name="event_log_{datetime}.log"
            )
            # general event logging -- all calls to __update__ will trigger this logger
            self._event_logger = EventLogger(logging_file)

    def add_task(
        self,
        name: str,
        path: str | list[str],
        agent_actuators: list[Callable[[], Actuator]] | None = None,
        avatar_actuators: list[Callable[[], Actuator]] | None = None,
        enable: bool = False,
    ):
        """Add a new task, this will load all required files and prepare the task but will not start the task (unless `enable` is True). To start the task use `enable_task` or have an agent take the `EnableTask` action.

        Args:
            name (str): the unique name of the task.
            path (str | list[str]): path(s) to task files.
            agent_actuators (list[Callable[[], Actuator]] | None, optional): actuators with which to create agents from a schedule file, see `MultiTaskEnvironment` documentation for details. Defaults to None.
            avatar_actuators (list[Callable[[], Actuator]] | None, optional): actuators that will be added to the avatar upon enabling the task, see `MultiTaskEnvironment` documentation for details. Defaults to None.
            enable (bool, optional): whether to immediately enable the task. Defaults to False.
        """
        if name in self._tasks:
            raise ValueError(f"Task with name {name} already exists.")
        if agent_actuators is None:
            agent_actuators = []
        if avatar_actuators is None:
            avatar_actuators = []
        self._task_loader.register_task(name, path)
        self._tasks[name] = self._task_loader.load(
            name, avatar_actuators=avatar_actuators, agent_actuators=agent_actuators
        )
        if enable:
            self.enable_task(name)

    def enable_task(
        self, task_name: str, context: dict[str, Any] | None = None, insert_at: int = -1
    ):
        """Manually enable a task. Tasks may otherwise be enabled via an `EnableTask` action.

        Args:
            task_name (str): name of the task to enable.
            context (dict[str,Any]): context to use when enabling the task
            insert_at (int): at what position in the SVG tree to insert the task element.
        """
        event = EnableTask(
            source=self.id, task_name=task_name, context=context, insert_at=insert_at
        )
        result = self.__update__(event)
        if isinstance(result, ErrorActiveObservation):
            raise result.exception()

    def disable_task(self, task_name: str):
        """Manually disable a task. Tasks may otherwise be disabled via an `EnableTask` action.

        Args:
            task_name (str): name of the task to disable.
        """
        event = DisableTask(source=self.id, task_name=task_name)
        result = self.__update__(event)
        if isinstance(result, ErrorActiveObservation):
            raise result.exception()

    def rename_task(self, task_name: str, new_name: str) -> None:
        """Rename a task. If the task is enabled this will alter the `id` attribute of the task element (TODO).

        Args:
            task_name (str): current name of the task.
            new_name (str): new name of the task.
        """
        if task_name in self._tasks:
            if new_name in self._tasks:
                raise ValueError(
                    f"Failed to rename task: {task_name}, another take with name: {new_name} already exists."
                )
            task = self._tasks[new_name]
            self._tasks[task_name] = task
            task.task_name = new_name
            del self._tasks[task_name]
            # TODO update the name of the task element if it has been enabled!
        else:
            raise ValueError(f"Failed to rename task: {task_name} as it doesn't exist.")

    def __update__(self, action: Event) -> ActiveObservation | ErrorActiveObservation:  # noqa
        # always log the action immediately (before execution)
        if self._event_logger:
            self._event_logger.log(action)
        # execute the action here or in super()
        if isinstance(action, EnableTask):
            return self._enable_task(action)
        elif isinstance(action, DisableTask):
            return self._disable_task(action)
        elif isinstance(action, INERT_ACTIONS):
            # these actions have no effect but are important for experiment logging
            return None
        else:
            return super().__update__(action)

    def on_user_input_event(  # noqa
        self,
        action: EyeMotionEvent
        | EyeMotionEventRaw
        | MouseButtonEvent
        | MouseMotionEvent
        | KeyEvent
        | WindowCloseEvent
        | WindowOpenEvent
        | WindowFocusEvent
        | WindowMoveEvent
        | WindowResizeEvent
        | ScreenSizeEvent,
    ):
        return super().on_user_input_event(action)

    def is_task_enabled(self, task_name: str) -> bool:
        """Is the given task enabled? Specially, is the task element part of the state? This will search for an element with `id` equal to the name of the task.

        Note that if the `id` of the task element has been changed elsewherethen this will not give the expected result. This will not happen with correct use of `MultiTaskAmbient`.

        Args:
            task_name (str): name of the task to check.

        Returns:
            bool: True if the task is enabled (is part of the state), False otherwise.
        """
        try:
            result = self._state.select(
                Select.new(f"/svg:svg/*[@id='{task_name}']", ["id"])
            )
        except XPathElementsNotFound:
            return False
        return result is not None and len(result) > 0

    @wrap_observation
    def _disable_task(self, event: DisableTask):
        """Disables a task - this will be called when a `DisableTask` event is received."""
        raise NotImplementedError("TODO")  # TODO

    @wrap_observation
    def _enable_task(
        self, event: EnableTask
    ) -> ActiveObservation | ErrorActiveObservation:
        """Enables a task - this will be called when an `EnableTask` event is received."""
        task = self._tasks[event.task_name]
        # TODO this assumes that the id of the task is "task_name"!
        if self.is_task_enabled(event.task_name):
            raise ValueError(f"Task {event.task_name} is already enabled.")
        agent = task.get_agent()
        if agent:
            self.add_agent(agent)
        avatar = task.get_avatar(self._avatar)
        # this should not change... the task should only add relevant actuators
        assert self._avatar == avatar
        # default is to insert as the first child of the root
        # TODO we need to check what the task 'id' is to keep track of which tasks are enabled and active!
        return self.__update__(
            Insert.new(
                xpath="/svg:svg",
                element=task.get_xml(event.context),
                index=event.insert_at,
            )
        )



"""Agent related functionality, including base classes for agents, actuators and sensors. Also contains some useful guidance related functionality, including some action definitions and `TaskAcceptabilitySensor`."""

from ..utils._task_loader import (
    agent_actuator,
    avatar_actuator,
)

from star_ray.agent import (
    Sensor,
    Actuator,
    Agent,
    attempt,
    observe,
)

from ..event.event_guidance import (
    ShowElementAction,
    HideElementAction,
    DrawArrowAction,
    DrawBoxAction,
    DrawBoxOnElementAction,
    DrawElementAction,
)
from .sensor_acceptability import (
    TaskAcceptabilitySensor,
    TaskAcceptabilityObservation,
)
from .agent_guidance import (
    GuidanceAgent,
)
from .sensor_userinput import (
    UserInputSensor,
)
from .actuator_guidance import (
    GuidanceActuator,
    CounterFactualGuidanceActuator,
    ArrowGuidanceActuator,
    BoxGuidanceActuator,
)
from .actuator_acceptability import (
    TaskAcceptabilityActuator,
)
from .actuator_avatar import AvatarActuator
from .avatar import Avatar

__all__ = (
    # decorators
    "agent_actuator",
    "avatar_actuator",
    "attempt",
    "observe",
    # type alias
    "Agent",
    "Sensor",
    "Actuator",
    # actions
    "ShowElementAction",
    "HideElementAction",
    "DrawArrowAction",
    "DrawBoxAction",
    "DrawBoxOnElementAction",
    "DrawElementAction",
    # sensors
    "UserInputSensor",
    "TaskAcceptabilitySensor",
    # actuators
    "GuidanceActuator",
    "TaskAcceptabilityActuator",
    "CounterFactualGuidanceActuator",
    "ArrowGuidanceActuator",
    "BoxGuidanceActuator",
    "AvatarActuator",
    # observations
    "TaskAcceptabilityObservation",
    # agents
    "GuidanceAgent",
    "Avatar",
)
